#include <string>
#include <vector>

#include "CBRFeature.h"

using namespace std;

CBRFeature::CBRFeature(std::string a) {
	name = a;
}

CBRFeature::CBRFeature() {}

int CBRFeature::getOffset(int a, int b) {
//    cout << "Offset med " << a << " " << b <<endl;
    // i must be smallest
    if(a > b) {
        int tmp = a;
        a = b;
        b = tmp;
    }
    
    int n = Value2Num.size();
	if(a >= n || b >= n) {
		cout << "A eller B out of bounds " << a << ","<<b<< "\n";
	}
    int sum = 0;
    for(int j = 0; j<a; j++) {
       sum += (n-1)-j;
    }
	sum +=(b-a)-1;
//	cout <<"Returnerer " << sum << endl;
    return sum;
}


// value holds the name of the feature value.
// XML does not necessary give the values in correct order
// store the value at place num in the 
void CBRFeature::addValue(std::string incoming, int num) {
	Value2Num[incoming] = num;
}


void CBRFeature::addRange(float fromin, float toin) {
	from = fromin;
	to = toin;
	cout << "Added "<< fromin<<" to "<<toin<<" to "<<name<<endl;
}


// get two strings, compute offset, store the float there
void CBRFeature::addSimilarity(std::string a, std::string b, float incoming) {

	unsigned int offset = getOffset(match(a), match(b));
	cout << "about to add similarity " << incoming <<  " between "<<a<<" and "<<b<< " at " <<offset<<endl;

	if(similarity.size() < offset+1) {
		similarity.resize(offset+1);
	}


	//Insert at right position, making vector n -> n+1.
	//delete the extra cell
	similarity.insert(similarity.begin() + offset, incoming);
	similarity.erase((similarity.begin() + offset + 1));
}

float CBRFeature::getNumericalSimilarity(float a, float b) {

	if(a == b) {
		return 1;
	}

	//look at the difference with respect to the allowed range of values.
	float diff = from - to;
	a -= from;
	b -= from;
	a /= diff;
	b /= diff;
	if(a >=b ) {
		return 1 - (a-b);
	}
	else {
		return 1 - (b-a);
	}
}


// return the similarity measure between to given feature values
float CBRFeature::getSimilarity(std::string a, std::string b) {


	//equal symbols == 1 in matching
	if(a.compare(b) == 0) {
		return 1.0;
	}
	int ai = match(a);
	int bi = match(b);
	if(ai == -1 || bi == -1) {
		cout << "Feil i getSimilarity " << a << "," << b << " :returnerte "<< ai <<","<< bi<<endl;
	}

	unsigned int off =  getOffset(ai,bi);

	//out of bounds? If offset > similarity.size then that value has not been defined.
	//if == 0.0 then not defined.
	if(similarity[off] == 0.0 || off >= similarity.size()) {
		cout << "No similarity defined for <" <<a<<","<<b<<">. Offset: "<<off<<" Using 0."<<endl;
		return 0;
	}
	return 1 - similarity[off];

}

int CBRFeature::match(std::string in) {

	//cout << "matching " << in <<" and allowed valued are: " <<endl;
	//PrintAllowedValues();

	//search vector for string in
	std::map<std::string, int>::const_iterator it = Value2Num.find(in);
	if (it != Value2Num.end())
		return it->second;
	return -1;
}


const std::string CBRFeature::getName() {
	return name;
}

const void CBRFeature::PrintAllowedValues() {
	map<std::string,int>::const_iterator iter;
	for(iter = Value2Num.begin(); iter != Value2Num.end(); ++iter ){
		cout << "Num:  "<< iter->second << " Name: "<< iter->first << endl;
	}
}


void CBRFeature::setNumeric(bool in) {
	numeric = in;
}

bool CBRFeature::isNumeric() {
	return numeric;
}


std::vector<std::string> CBRFeature::getAllowedValues() {
	vector<string> ret;
	
	//populate vector
	map<string,int>::const_iterator iter;
	for(iter = Value2Num.begin(); iter != Value2Num.end(); ++iter ){
		ret.push_back(iter->first);
	}
	return ret;
}