#ifndef CBR_I_F
#define CBR_I_F

#include <itkImage.h>
#include <climits>
#include <vector>
#include <utility>
#include <itkRegionOfInterestImageFilter.h>
#include <itkNeighborhoodConnectedImageFilter.h>
#include <itkAddImageFilter.h>
#include <itkConnectedComponentImageFilter.h>
#include <itkRelabelComponentImageFilter.h>
#include <itkScalarImageToHistogramGenerator.h>
#include <itkCastImageFilter.h>
#include <itkSmoothingRecursiveGaussianImageFilter.h>
#include <itkReinitializeLevelSetImageFilter.h>

#include <fstream>

#include "CBRCase.h"
#include "CBRStatistics.h"


template<class ImageType>
typename ImageType::Pointer GetArrivedFile(CBRCase from, CBRCase to, std::string fileName) {
	std::string cat = "out\\Revise3\\"+from.GetStringIndex("file")+"\\"+to.GetStringIndex("file")+"\\"+fileName;
	std::cout << "Returnerer : " << cat << std::endl;
	typedef itk::Image<unsigned int, 3> InnType;
	typedef itk::CastImageFilter<InnType, ImageType> CasterType;
	CasterType::Pointer caster = CasterType::New();

	//get file and cast to what we want.
	InnType::Pointer inn = GetImageFile<InnType>(cat);
	caster->SetInput(inn);
	ImageType::Pointer ret = caster->GetOutput();
	ret->Update();
	ImageType::Pointer smallRet = GetSubvolumeFromCase<ImageType>(ret, ret, from).first;
	smallRet->Update();
	return smallRet;
}


template<class T>
std::pair<typename T::IndexType, typename T::IndexType> getVentricleBoundary(typename T::Pointer solution, int value=VentricleValue){

	//return values
	T::IndexType min;
	T::IndexType max;



	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;
	ConstIteratorType it(solution, solution->GetRequestedRegion());

	// variables to hold maxima
	int xmin = std::numeric_limits<int>::max();
	int xmax = 0;
	int ymin = std::numeric_limits<int>::max();
	int ymax = 0;
	int zmin = std::numeric_limits<int>::max();
	int zmax = 0;

	// get coordinates of non-zero pixel

	T::IndexType index;
	//go through iterator and get top, bottom,left,right,front and rear maxima	
	it.GoToBegin();
	while(!it.IsAtEnd()) {
		if(it.Get() == value ) {
			index = it.GetIndex();
			// x
			if(index[0] < xmin) xmin = index[0];
			if(index[0] > xmax) xmax = index[0];
			// y
			if(index[1] < ymin) ymin = index[1];
			if(index[1] > ymax) ymax = index[1];
			// z
			if(index[2] < zmin) zmin = index[2];
			if(index[2] > zmax) zmax = index[2];

		}
		++it;
	}
	
	min[0] = xmin;
	min[1] = ymin;
	min[2] = zmin;

	max[0] = xmax;
	max[1] = ymax;
	max[2] = zmax;

	return std::make_pair(min, max);

}

template<class T>
typename T::Pointer getSubvolume(typename T::Pointer image, typename T::IndexType &min, typename T::IndexType &max) {
		/*
		This returns the region from an image specified by coords min and max
	*/

	T::SizeType size2 = GetSizeOfImage<T>(image);
//	std::cout << "Str i getSubvolume: "<< size2 << std::endl;
	T::RegionType region;

	typedef itk::RegionOfInterestImageFilter<T,T> FilterType;
	FilterType::Pointer filter = FilterType::New();

	//compute size in each direction
	T::SizeType size;
	size[0] = max[0] - min[0];
	size[1] = max[1] - min[1];
	size[2] = max[2] - min[2];
	
//	std::cout << size << "," << max << std::endl;

	// set size and origin
	region.SetSize(size);
	region.SetIndex(min);

//	std::cout << "Size: " << size << std::endl;

	//set up filter with this region of interest
	filter->SetRegionOfInterest(region);
	filter->SetInput(image);

	T::Pointer retImage = filter->GetOutput();

	// update image to process pipeline and get result.
	try{
		retImage->Update();
	}
	catch(itk::ExceptionObject &exp) {
		std::cerr << exp;
	}

	return retImage;
}
//mode is either "ventricle" or anything else.
//if ventricle the min/max of the area of ventricle is used. 
//else, vica versa
template<class T>
void getSubvolumeMinMax(typename T::Pointer orig, typename T::Pointer seg, std::string mode = "subvolume") {
	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;

	
	// set up iterators
	ConstIteratorType solIt(seg, seg->GetRequestedRegion());
	ConstIteratorType origIt(orig, orig->GetRequestedRegion());

	int max = 0;
	int min = std::numeric_limits<int>::max();

	solIt.GoToBegin();
	origIt.GoToBegin();
	while(!origIt.IsAtEnd()) {
		if(mode == "ventricle") {
			if(solIt.Get() == VentricleValue){
				std::cout << origIt.GetIndex() << std::endl;
				if(origIt.Get() > max)
					max = origIt.Get();
				if(origIt.Get() < min)
					min = origIt.Get();
			}
		}
		else {
			if(solIt.Get() != VentricleValue){
					if(origIt.Get() > max)
					max = origIt.Get();
				if(origIt.Get() < min)
					min = origIt.Get();
			}
		}
		++solIt;
		++origIt;
	}

	std::cout <<mode <<"- min: " << min << " max: " << max << std::endl;

//	GetHistogramFromImage<T>(orig, 255, 10.0);
}



template<class T>
typename T::Pointer Threshold(typename T::Pointer image, int thr) {
	typedef itk::ImageRegionIterator<T> IteratorType;
	std::cout << "Thresholding at " << thr << std::endl;
	T::Pointer retImage = CBRCopyImage<T>(image);
	
	// set up iterators
	IteratorType it(retImage, retImage->GetLargestPossibleRegion());

	it.GoToBegin();
	while(!it.IsAtEnd()) {
		if(it.Get() <= thr) {
			it.Set(255);
//			std::cout << "X";
		}
		else {
			it.Set(0);
//			std::cout << std::endl<< std::endl<< std::endl<< std::endl;
		}
		++it;
	}
	return retImage;
}



template<class T>
typename T::SizeType GetSizeOfImage(typename T::Pointer image) {

	typedef itk::ImageRegion<Dimension> RegionType;
	RegionType region = image->GetLargestPossibleRegion();	
	RegionType::SizeType size = region.GetSize();

	return size;
}

/*

	Computes the score between the proposed solution and the solution.
	Intersection(A,B) / Union(A,B)
	
*/
template<class ImageType>
float getScore(typename ImageType::Pointer thresh, typename ImageType::Pointer solution, int lowerThres=0, int finalThres=0) {

	typedef itk::ImageRegionConstIterator<ImageType> ConstIteratorType;
	typedef itk::ImageRegionIterator<ImageType > IteratorType;

	IteratorType it(thresh, thresh->GetLargestPossibleRegion());
	ConstIteratorType itSol(solution, solution->GetLargestPossibleRegion());
	float score = 0;
	float unionn = 0;
	float intersect = 0;
	it.GoToBegin();
	itSol.GoToBegin();
	while(!it.IsAtEnd()) {

		if(itSol.Get() == VentricleValue) {
			if(it.Get() == 255) {  // direct match
				intersect++;
			}
			else {
				unionn++;
			}
		}
		else {
			if(it.Get() == 255) {
				unionn++;
			}
		}
		++it;
		++itSol;
	}

	unionn += intersect;
	if(unionn == 0) {
		std::cout << "score: 0 "<< std::endl;
	}
	else {
		score = intersect / unionn;
		std::cout << finalThres << " - " << lowerThres << " " << "Score:  " << score << " \tIntersection/Union: " << intersect << "/" << unionn <<std::endl;
	}
	return score;
}






template<class T>
typename T::Pointer createMatchImageToImage(typename T::Pointer image, typename T::Pointer solImage) {

	//	std::cout << "Matching solutions " <<std::endl;
	//iterators
	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;
	typedef itk::ImageRegionIterator<T > IteratorType;

	T::Pointer retImage = CBRCopyImage<T>(solImage);

	T::Pointer mImage = image;
	solImage->Update();
	mImage->Update();




	//create iterators.
	ConstIteratorType it(image, image->GetLargestPossibleRegion());
	ConstIteratorType itS(solImage, solImage->GetLargestPossibleRegion());
	IteratorType itM(retImage, retImage->GetLargestPossibleRegion());

	// put together overlap-image.
	int corr = 0;
	int fp = 0;
	int fn = 0;
	itS.GoToBegin();
	itM.GoToBegin();
	it.GoToBegin();
	while(!it.IsAtEnd()) {
		int fVal = it.Get();
		int sVal = itS.Get();
		if(fVal == 255) {
			if(sVal == VentricleValue) {
				//direct match
				itM.Set(255);
				corr++;
			}
			else {
				//false positive
				itM.Set(100);
				fp++;
			}
		}
		// false negative
		else {
			if(sVal == VentricleValue) {
				itM.Set(150);
				fn++;
			}
			else {
				// no match
				itM.Set(0);
			}
		}
		++itS;
		++itM;
		++it;
	}

	return retImage;
}

std::string ToString(float in) {
	std::ostringstream os;
	os << in;
	std::string ret = os.str();
	return ret;
}



template<class T>
typename T::Pointer CopySolutionImageToImage(typename T::Pointer image, typename T::Pointer solImage){

	unsigned int count = 0; 
	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;
	typedef itk::ImageRegionIterator<T > IteratorType;

	IteratorType it(image, image->GetLargestPossibleRegion());
	ConstIteratorType itSol(solImage, solImage->GetLargestPossibleRegion());


	it.GoToBegin();
	itSol.GoToBegin();
	while(!it.IsAtEnd()) {
		if(itSol.Get() == VentricleValue) {
			it.Set(255);
			count++;
		}
		else
			it.Set(0);
		++it;
		++itSol;
	}

	std::cout << "Solution copied, " << count << " pixels copied." <<std::endl;

//	WriteImageFile<T>(image, "d:\\diplom\\Diplom\\code\\out\\copySolution.mhd");
	return image;
}


template<class T>
typename T::Pointer DivideInEight(typename T::Pointer image, typename T::Pointer solImage){

	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;
	typedef itk::ImageRegionIterator<T > IteratorType;

	IteratorType it(image, image->GetRequestedRegion());
	ConstIteratorType itSol(solImage, solImage->GetRequestedRegion());

	T::SizeType size = GetSizeOfImage<T>(image);
	int xmidt = (int)(size[0] / 2);
	int ymidt = (int)(size[1] / 2);
	int zmidt = (int)(size[2] / 2);

	T::IndexType index;
	it.GoToBegin();
	itSol.GoToBegin();
	while(!it.IsAtEnd()) {
		index = it.GetIndex();
		//if(itSol.Get() == VentricleValue) {
			if(index[0] == 0 || index[0] == xmidt || index[0] == (size[0]-1)) it.Set(200);
			if(index[1] == 0 || index[1] == ymidt || index[1] == (size[1]-1)) it.Set(200);
			if(index[2] == 0 || index[2] == zmidt || index[2] == (size[2]-1)) it.Set(200);
//			it.Set(255);
//		}
		//else {
		//it.Set(0);
			//if(index[0] == 0 || index[0] == xmidt || index[0] == (size[0]-1)) it.Set(200);
			//if(index[1] == 0 || index[1] == ymidt || index[1] == (size[1]-1)) it.Set(200);
			//if(index[2] == 0 || index[2] == zmidt || index[2] == (size[2]-1)) it.Set(200);
		//}
		++it;
		++itSol;
	}


//	WriteImageFile<T>(image, "d:\\diplom\\Diplom\\code\\out\\copySolution.mhd");
	return image;
}


template<class ImageType>
typename ImageType::Pointer blackOut(typename ImageType::Pointer original, int thr) {

	// get a copy of the original image.
	//itk::CastImageFilter<ImageType, ImageType>::Pointer caster = itk::CastImageFilter<ImageType,ImageType>::New();
	//caster->SetInput(original);
	ImageType::Pointer image = CBRCopyImage<ImageType>(original);
//	image->Update();

//	WriteImageFile<ImageType>(image, "out\\revise\\image-"+ToString(thr)+".mhd");

	//first, threshold the incoming image with spesified threshold
	image = Threshold<ImageType>(image, thr);

//	WriteImageFile<ImageType>(image, "out\\revise\\threshed-"+ToString(thr)+".mhd");

	// work on the threholded image and find unwanted pixels.
	// also get number of regions
	int regionCount = 0;
	image = colorRegionBySize<ImageType>(image, &regionCount);

//	WriteImageFile<ImageType>(image, "out\\revise\\blackout-"+ToString(thr)+".mhd");
	
//	exit(1);

	// get histogram to get region color with most pixels. the one to delete	
	typedef itk::Statistics::ScalarImageToHistogramGenerator<ImageType> histTypeGenerator;
	typedef histTypeGenerator::HistogramType histType;

	histTypeGenerator::Pointer histGen = histTypeGenerator::New();
	
//	std::cout << "Region count " <<regionCount << std::endl;

	histGen->SetInput(image);

	histGen->SetNumberOfBins(  1000 );
	histGen->SetMarginalScale(    1 );
	histGen->SetHistogramMin(     1 );
	histGen->SetHistogramMax(  1000 );
	histGen->Compute();

	const histType * histogram = histGen->GetOutput();

//	std::cout << "Got histogram" << std::cout;


	//iterate to find biggest region. actually sorted to 1, it seems.
	histType::ConstIterator it = histogram->Begin();
	histType::ConstIterator end = histogram->End();
	int color = 0;
	int biggestRegion = 0;
	int counter = 1;
	while(it != end) {
//		std::cout << "color: " << counter << " freq: " << it.GetFrequency() << std::endl;
		if( it.GetFrequency() > biggestRegion) {
			biggestRegion = it.GetFrequency();
//			std::cout << "Biggest region: " << it.GetFrequency() << std::endl;
			color = counter;
		}
		counter++;
		++it;
	}

	// get histogram end

// set color from histogram max to 255, will not be thresholded in later thresholding
	typedef itk::ImageRegionIterator<ImageType> ittype;
	ittype itBlack(image, image->GetLargestPossibleRegion());

	// iterate through original image
	ittype itRet(original, original->GetLargestPossibleRegion());

	itBlack.GoToBegin();
	itRet.GoToBegin();
	int asdf = 0;
	// iterate through original and image colorized by size.
	// set all pixels to 255 in original if they are 1 in colorised.
	while(!itBlack.IsAtEnd()) {

		if(itBlack.Get() == 1) {
			itRet.Set(255);
			asdf++;
		}
		//else
		//	itRet.Set(0);
		++itBlack;
		++itRet;
	}

	std::cout << "Fjernet: "<< asdf << std::endl;
//	WriteImageFile<ImageType>(original, "blackedOut-"+ToString(thr)+".mhd");

	return original;
}


// sets all other regions than the "regions" number biggest to 0.
template<class ImageType>
typename ImageType::Pointer keepBiggestRegions(typename ImageType::Pointer image, int regions) {
	
	// colorize regions according to size 
	int unused = 0;
	ImageType::Pointer retImage = CBRCopyImage<ImageType>(image);
	retImage = colorRegionBySize<ImageType>(retImage, &unused);

	itk::ImageRegionIterator<ImageType> it(retImage, retImage->GetRequestedRegion());

	it.GoToBegin();
	while(!it.IsAtEnd()) {
		if( it.Get() > regions ) {	
			it.Set(0);
		}
		++it;
	}

	// colorRegionBySize has the remaining values from 1...regions.
	// since it is the result of segmentation, set to 255 again.
	retImage = setRegionsToWhite<ImageType>(retImage);
	return retImage;
}

//	blacked = keepSmallestRegions<ImageType>(blacked, 2);
// sets all other regions than the "regions" number biggest to 0.
template<class ImageType>
typename ImageType::Pointer keepRegion(typename ImageType::Pointer image, std::vector<int> allowed) {
	
	// colorize regions according to size 
	int unused = 0;
	image = colorRegionBySize<ImageType>(image, &unused);

	itk::ImageRegionIterator<ImageType> it(image, image->GetRequestedRegion());

	it.GoToBegin();
	while(!it.IsAtEnd()) {
		if(notContains(it.Get(), allowed)) {	
			it.Set(0);
		}
		++it;
	}

	setRegionsToWhite<ImageType>(image);
//	used to debug
//	WriteImageFile<ImageType>(image, "onlyOne.mhd");
	return image;
}

template<class ImageType>
typename ImageType::Pointer colorRegionBySize(typename ImageType::Pointer image, int *regionCount) {

	typedef itk::ConnectedComponentImageFilter<ImageType, ImageType> ConnectType;
	typedef itk::RelabelComponentImageFilter<ImageType, ImageType> RelabelType;
	ConnectType::Pointer connector = ConnectType::New();
	RelabelType::Pointer relabler = RelabelType::New();

	connector->SetInput(image);
	relabler->SetInput(connector->GetOutput());
	ImageType::Pointer retur = relabler->GetOutput();

	//save this number.
	*regionCount = relabler->GetNumberOfObjects();

	try{
		retur->Update();
	}
	catch(itk::ExceptionObject &exp) {
		std::cerr << exp;
		exit(1);
	}

	return retur;
}


//after image is colorized and only x is kept, set the remaining regions to white again.
template<class T>
typename T::Pointer setRegionsToWhite(typename T::Pointer image) {
	itk::ImageRegionIterator<T> it(image, image->GetRequestedRegion());

	it.GoToBegin();
	while(!it.IsAtEnd()) {
		if(it.Get() > 0)
			it.Set(255);
		++it;
	}

	return image;
}

template<class T>
void printVentricleBoundary(typename T::Pointer solutionFS) {
	
	std::pair<T::IndexType, T::IndexType> minmax;
	minmax = getVentricleBoundary<T>(solutionFS);
	std::cout << "min: " << minmax.first << " max " << minmax.second << std::endl;
	exit(1);
}




/*

	Take the solution and make an image with gradients used to guide the level set.

*/
template<class ImageType>
typename ImageType::Pointer createFastMarchingImage(typename ImageType::Pointer image) {
	
	//smooth image to get smoother transitions. 
	typedef itk::SmoothingRecursiveGaussianImageFilter<ImageType, ImageType> SmoothType;
	SmoothType::Pointer smoother = SmoothType::New();
	smoother->SetSigma(1.0);

	//get a copy of the input image. Little trick.
	itk::CastImageFilter<ImageType, ImageType>::Pointer caster = itk::CastImageFilter<ImageType, ImageType>::New();
	caster->SetInput(image);
	ImageType::Pointer returnImg = caster->GetOutput();
	returnImg->Update();
	
	typedef itk::ImageRegionIterator<ImageType> itType;
	itType it(returnImg, returnImg->GetRequestedRegion());
	
	while(!it.IsAtEnd()) {
		if(it.Get() == VentricleValue)
			it.Set(255.0);
		else
			it.Set(0.0);
	++it;
	}

	
//typedef itk::ReinitializeLevelSetImageFilter<ImageType> DistanceFilterType;
//DistanceFilterType::Pointer m_DistanceFilter = DistanceFilterType::New();
//   m_DistanceFilter->SetInput( returnImg );
//   m_DistanceFilter->NarrowBandingOn();
//   m_DistanceFilter->SetNarrowBandwidth( 10 );
//
//   m_DistanceFilter->Update();

	//smooth image. 


	return returnImg;
}




/*
	Method to adapt the parameters of a CBRCase to better fit the incoming case

*/
void Adapt(CBRCase &oldCase, CBRCase &newCase) {
	oldCase.caseCopy(newCase);
	std::cout << "Information copied to new case:" << std::endl;
	newCase.Print();
}




void AdaptAverage(CBRCase &oldCase, CBRCase &newCase) {
	newCase.SetIntIndex("Xmax", oldCase.GetIntIndex("Xmax"));
	newCase.SetIntIndex("Ymax", oldCase.GetIntIndex("Ymax"));
	newCase.SetIntIndex("Zmax", oldCase.GetIntIndex("Zmax"));
	newCase.SetIntIndex("Xmin", oldCase.GetIntIndex("Xmin"));
	newCase.SetIntIndex("Ymin", oldCase.GetIntIndex("Ymin"));
	newCase.SetIntIndex("Zmin", oldCase.GetIntIndex("Zmin"));
	
	//comput average
	typedef itk::Image<int, 3> ImageType;
	std::string file = "orig\\" + newCase.GetStringIndex("file") + ".hdr"; 
	ImageType::Pointer tmpImage = GetImageFile<ImageType>(file);
	ImageType::IndexType min; min[0] = oldCase.GetIntIndex("Xmin"), min[1] = oldCase.GetIntIndex("Ymin"), min[2] = oldCase.GetIntIndex("Zmin");
	ImageType::IndexType max; max[0] = oldCase.GetIntIndex("Xmax"), max[1] = oldCase.GetIntIndex("Ymax"), max[2] = oldCase.GetIntIndex("Zmax"); 
	ImageType::Pointer subImg = getSubvolume<ImageType>(tmpImage, min, max);
	std::vector<float> histogramStatistics= GetHistogramStatistics<ImageType>(subImg);

	// use average to offset thresholds
	int diff = oldCase.GetIntIndex("average") - (int)histogramStatistics[0];
	std::cout << oldCase.GetIntIndex("average") << "-" << (int)histogramStatistics[0] << "=" << diff << std::endl;
	int first = oldCase.GetIntIndex("firstThresh") - diff;
	int second = oldCase.GetIntIndex("secondThresh") - diff;
	if(first < 0 ) first = 0;
	if(second < 0) second = 0;
	newCase.SetIntIndex("firstThresh", first);
	newCase.SetIntIndex("secondThresh", second);

	std::cout << "Endret frste fra " << oldCase.GetIntIndex("firstThresh") << " til " << first << std::endl;
	std::cout << "Endret andre fra " << oldCase.GetIntIndex("secondThresh") << " til " << second << std::endl;

}


/************************************************************************

right now this is used to try out different stuff to adapt cases. 

************************************************************************/
template<class ImageType>
void AdaptDebug(typename ImageType::Pointer image, typename ImageType::Pointer fullImage) {

	// a vector containing the statistics on the histogram
	//std::vector<float> histogramStatistics= GetHistogramStatistics<ImageType>(fullImage);
	//std::cout << std::endl;
	//std::cout << "<AverageFull>" << histogramStatistics[0] << "</AverageFull>" << std::endl;
	//std::cout << "<VarianceFull>" << histogramStatistics[1] << "</VarianceFull>" << std::endl;
	//std::cout << "<SkewnessFull>" << histogramStatistics[2] << "</SkewnessFull>" << std::endl;
	//std::cout << "<KurtosisFull>" << histogramStatistics[3] << "</KurtosisFull>" << std::endl;
	//std::cout << "<CoeffVariationFull>" << histogramStatistics[4] << "</CoeffVariationFull>" << std::endl;
	//std::cout << "<EntropyFull>" << histogramStatistics[5] << "</EntropyFull>" << std::endl;
	//std::cout << "<CentroidXFull>" << histogramStatistics[6] << "</CentroidXFull>" <<std::endl;
	//std::cout << "<CentroidYFull>" << histogramStatistics[7] << "</CentroidYFull>" << std::endl;
	//std::cout << "<CentroidZFull>" << histogramStatistics[8] << "</CentroidZFull>" << std::endl << std::endl;

	std::vector<float> histogramStatistics= GetHistogramStatistics<ImageType>(image);
	std::cout << std::endl;
	std::cout << "<Average>" << histogramStatistics[0] << "</Average>" << std::endl;
	std::cout << "<Variance>" << histogramStatistics[1] << "</Variance>" << std::endl;
	std::cout << "<Skewness>" << histogramStatistics[2] << "</Skewness>" << std::endl;
	std::cout << "<Kurtosis>" << histogramStatistics[3] << "</Kurtosis>" << std::endl;
	std::cout << "<CoeffVariation>" << histogramStatistics[4] << "</CoeffVariation>" << std::endl;
	std::cout << "<Entropy>" << histogramStatistics[5] << "</Entropy>" << std::endl;
	std::cout << "<CentroidX>" << histogramStatistics[6] << "</CentroidX>" <<std::endl;
	std::cout << "<CentroidY>" << histogramStatistics[7] << "</CentroidY>" << std::endl;
	std::cout << "<CentroidZ>" << histogramStatistics[8] << "</CentroidZ>" << std::endl << std::endl;

}




/************************************************************************

	Write histogram to disk. Used for debugging.

************************************************************************/
template<class histType>
void WriteHistogramToDisk(const histType *histogram) { 
	
	histType::ConstIterator it = histogram->Begin();
	histType::ConstIterator end = histogram->End();

	// open file for write
	ofstream file("text\\histogram.txt");

	// write
	while(it != end) {
		file << it.GetFrequency() << std::endl;
		++it;
	}
	std::cout << "finished printing" << std::endl << std::flush;

	// close handle
	file.close();
	
}



/**************************************************************************************

	Casts the incoming image from one template to another.
	Used primarily to cast bestImage from float to int in UseThreshold
	

**************************************************************************************/
template<class From, class To>
typename itk::Image<To, 3>::Pointer CastImage(typename itk::Image<From,3>::Pointer bestImg) {
	typedef itk::Image<From, 3> OldImageType;
	typedef itk::Image<To, 3> NewImageType;

	typedef itk::CastImageFilter<OldImageType, NewImageType> CasterType;
	CasterType::Pointer caster = CasterType::New();

	caster->SetInput(bestImg);

	NewImageType::Pointer retImg = caster->GetOutput();

	try{
		retImg->Update();
	}
	catch(itk::ExceptionObject &exp) {
		std::cerr << exp << std::endl;
	}


	return retImg;

}




/************************************************************************************

	Merge the subimage into the large image
	Iterate and use get/set pixel at correct index.

************************************************************************************/

template<class ImageType>
void MergeImages(typename ImageType::Pointer orig, typename ImageType::Pointer done, typename ImageType::IndexType min) {

	ImageType::SizeType size = GetSizeOfImage<ImageType>(done);
	ImageType::IndexType BigIndex;

	typedef itk::ImageRegionIterator<ImageType> itType;
	itType it(done, done->GetRequestedRegion());

	// index in bigimage bottomleft coordinate in small image (0,0,0) pluss offset this is in the big image (min)
	// I think :)

	while(!it.IsAtEnd()) {
		ImageType::IndexType SmallIndex = it.GetIndex();
		BigIndex[0] = min[0] + SmallIndex[0];
		BigIndex[1] = min[1] + SmallIndex[1];
		BigIndex[2] = min[2] + SmallIndex[2];
		if(it.Get() == 255)
			orig->SetPixel(BigIndex, it.Get());
		++it;
	}
}



/************************************************************************************************

	Get size of regions

************************************************************************************************/
template<class ImageType>
std::vector<unsigned int> GetSizeOfRegions(typename ImageType::Pointer image) {
	std::vector<unsigned int> retVector;
	int numRegions=-1;
	image->Update();
//	typename ImageType::Pointer colorRegionBySize(typename ImageType::Pointer image, int *regionCount) {
	ImageType::Pointer coloredImg = colorRegionBySize<ImageType>(image, &numRegions);

	typedef itk::ImageRegionConstIterator<ImageType> ItType;
	ItType it(coloredImg, coloredImg->GetLargestPossibleRegion());
//	std::cout << "Number: " << numRegions << std::endl;


	retVector.push_back(0);
	retVector.push_back(0);
	it.GoToBegin();
	while(!it.IsAtEnd()) {
		if(it.Get() == 1)
			retVector[0]++;
		if(it.Get() == 2)
			retVector[1]++;
		++it;
	}
//	std::cout << "Region 1,2 har " << retVector[0] << "," << retVector[1] << " pixler " << std::endl;

	return retVector;	 
}


template<class ImageType> 
std::pair<typename ImageType::Pointer, typename ImageType::Pointer> GetSubvolumeFromCase(typename ImageType::Pointer orig, typename ImageType::Pointer seg, CBRCase incase) {

	ImageType::IndexType min;
	ImageType::IndexType max;

	min[0] = incase.GetIntIndex("Xmin");
	min[1] = incase.GetIntIndex("Ymin");
	min[2] = incase.GetIntIndex("Zmin");
	max[0] = incase.GetIntIndex("Xmax");
	max[1] = incase.GetIntIndex("Ymax");
	max[2] = incase.GetIntIndex("Zmax");

	// get subvolume. Save to disk for inspection.
	ImageType::Pointer subimage = getSubvolume<ImageType>(orig, min, max);
	ImageType::Pointer subsol = getSubvolume<ImageType>(seg, min, max);
	std::pair<typename ImageType::Pointer, typename ImageType::Pointer> retPair;
	retPair.first = subimage;
	retPair.second = subsol;

	return retPair;
	
}

/*******************************************************************************************************************

	Returns the maximum and minimum value for this attribute

*******************************************************************************************************************/
std::pair<float, float> GetMinMaxForAttribute(std::string atrName, std::vector<CBRCase> cases) {
	std::pair<float, float> minmax; 
	minmax.first = std::numeric_limits<float>::max();
	minmax.second = std::numeric_limits<float>::min();

	for(unsigned int i = 0; i < cases.size() ; i++) {
		float tmp = cases[i].GetFloatIndex(atrName);
		if(tmp < minmax.first) minmax.first = tmp;
		if(tmp > minmax.second) minmax.second = tmp;
	}
//	std::cout << "Minmax found to be: " << minmax.first << "," << minmax.second << std::endl;
	return minmax;
}

/********************************************************************************************************************

Computes the distance between two cases, equation from Perner

********************************************************************************************************************/
float GetStatisticalDistance(float a, float b, float min, float max) {
	float ret = ( ( a-min )/( max - min ) ) - ( ( b - min )/( max - min) );
	if(ret < 0)
		ret *= -1;
//	std::cout << a << "-" << b <<"/" << max << "-" << min << "=" << ret << std::endl;
	return ret;
}


/******************************************************************************************************************

	Computes the statistics for the image in the new case

******************************************************************************************************************/
std::vector<float> GetStatisticsFromNewImage(std::string filename) {
	typedef itk::Image<float, 3> ImageType;
	std::cout << std::endl;
	std::string file = "orig\\" + filename + ".hdr";
	ImageType::Pointer image = GetImageFile<ImageType>(file);
	std::vector<float> stats = GetHistogramStatistics<ImageType>(image);
	return stats;
}

/**************

			Makes a copy on an image, do not work on same image as things are iterated.

**************/
template<class ImageType>
typename ImageType::Pointer CBRCopyImage(typename ImageType::Pointer original) {
	// get a copy of the original image.
	itk::CastImageFilter<ImageType, ImageType>::Pointer caster = itk::CastImageFilter<ImageType,ImageType>::New();
	caster->SetInput(original);
	ImageType::Pointer image = caster->GetOutput();
	image->Update();
	return image;
}



/*********************************************************************************************************

	Critize the solution
	The incomingCase has a priori knowledge that is to be used in validation.
	If not OK is reports a failure and an error code back to Revise() that will handle it

*********************************************************************************************************/
template<class ImageType>
int Judge(typename ImageType::Pointer image, CBRCase &incomingCase) {
	ImageType::IndexType centroid = GetCentroids<ImageType>(image);
	std::cout << "Incoming: [" << incomingCase.GetIntIndex("RegionX") << "," << incomingCase.GetIntIndex("RegionY") << "," << incomingCase.GetIntIndex("RegionZ") <<"] vs ";
	std::cout << "Computed Centroid: " << centroid << std::endl;
	
	int returnCode = 1;
	// is the region found the one that should have been removed?
	returnCode = CheckForFirstRegion<ImageType>(image, incomingCase, centroid);
	if(returnCode == 2)
		return 2;

	// did we find a completely wrong region?
	returnCode = CheckForWrongRegion<ImageType>(image, incomingCase, centroid);
	if(returnCode > 1)
		return returnCode;



	// everything is OK. 
	return 1;
}




/******************************************************************************************************************************

			Revise - step. Flags an error if the segment found is wrong

******************************************************************************************************************************/
template<class ImageType>
int Revise(typename ImageType::Pointer image, CBRCase &incomingCase, int num) {

	int errorNo = Judge<ImageType>(image,incomingCase);
	
	WriteImageFile<ImageType>(image, "out\\revise\\reviseImg-"+ToString(num)+".mhd");

	switch(errorNo) {
		case 2: 	// found region is the one that should be eliminated. 
					// and ventricles not removed
			incomingCase.SetIntIndex("firstThresh", incomingCase.GetIntIndex("firstThresh") + 5); 
//			incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("firstThresh") + 5); 
			break;

		case 3: 
			incomingCase.SetIntIndex("firstThresh", incomingCase.GetIntIndex("firstThresh") - 5);
			//incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("firstThresh") + 5); 
			break;

		case 4: 
			incomingCase.SetIntIndex("firstThresh", incomingCase.GetIntIndex("firstThresh") - 5);
			//incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("firstThresh") + 5); 
			break;
	}

	return errorNo;
}






/************************************************************************************************************

			Tries to determine if the region found is the one that should be removed with the first threshold.
			Returns 2 if it suspects/detects a failure and 1 elsewise

*************************************************************************************************************/
template<class ImageType>
int CheckForFirstRegion(typename ImageType::Pointer image, CBRCase &incomingCase, typename ImageType::IndexType centroid) {
	int xdiff = abs(centroid[0] - incomingCase.GetIntIndex("RegionX"));
	if((abs(centroid[2] - incomingCase.GetIntIndex("RegionZ")) > 8 || (centroid[2] < incomingCase.GetIntIndex("RegionZ") * 0.5))
		&& ( xdiff < 10)) {
			std::cout << std::endl <<  "Failed on centroid-check.";
			std::cout << "Klassifisert som region som skal fjernes." << std::endl;
			// er det fordi ventriklene er fjernet tidligere, eller for lav frste threshold?
			// 2 = ventriklene er fjernet, 3 = de er ikke det.
			return CheckForWhiteOut<ImageType>(image, incomingCase, centroid);
		}
		return 1;
}


/************************************************************************************************************

			Tries to determine if the region found is a wrong one because the ventricles were removed in the first step.
			Returns 3 if it suspects/detects a failure and 1 elsewise

*************************************************************************************************************/
template<class ImageType>
int CheckForWrongRegion(typename ImageType::Pointer image, CBRCase &incomingCase, typename ImageType::IndexType centroid) {
	int size = GetSizeOfRegions<ImageType>(image)[0];
	int wantedSize = incomingCase.GetIntIndex("size");
	// too small a region to be correct?
	//if(size < wantedSize/4) {
	//	std::cout << std::endl << "Mistenker for liten/helt feil region" << std::endl;
	//	return 3;
	//}

	// way of to one side?
	int xdiff = abs(centroid[0] - incomingCase.GetIntIndex("RegionX"));
	if(xdiff > 15){
		std::cout << std::endl << "Mistenker helt feil region" << std::endl;
		return CheckForWhiteOut<ImageType>(image, incomingCase, centroid);
	}

	//way of to another side?
	int ydiff = abs(centroid[1] - incomingCase.GetIntIndex("RegionY"));
//TODO: DENNE FUNKER IKKE.
	if((centroid[1] < (incomingCase.GetIntIndex("RegionY")*.5)) || ydiff > 10){
		std::cout << std::endl <<  "Mistenker feil region funnet.";
		std::cout << "Feilet p Y-koord sjekk. - senker frste thresh." << std::endl;
		return 4;
	}
	return 1;
}

template<class ImageType>
int CheckForWhiteOut(typename ImageType::Pointer image, CBRCase &incomingCase, typename ImageType::IndexType centroid) {
	std::cout << "Checking if whiteout has occured" << std::endl;
	
	int size = GetSizeOfRegions<ImageType>(image)[0];
	int wantedSize = incomingCase.GetIntIndex("size");
	 //too small a region to be correct?
	if(size > wantedSize*2) {
		std::cout << std::endl << "Whiteout detected - lower first threshold" << std::endl;
		return 3;
	}
	std::cout << "White out not detected - increase first treshold" << std::endl;
	return 2;


}




/********************************************************************************************************************

			SECOND ATTEMPT AT REVISE AND REPAIR BELOW

********************************************************************************************************************/


template<class ImageType>
int ReviseSecondAttempt(typename ImageType::Pointer image, CBRCase &incomingCase, int num, typename ImageType::Pointer sol, std::string cat) {
	int ret = 0, mode = incomingCase.GetIntIndex("mode");

	// gather debug stats
	std::vector<float> mom = GetMoments<ImageType>(image);
	std::cout << "Momenter: " << mom[0] << ", " << mom[1] << ", " << mom[2];
	std::cout << " vs wanted: "<< incomingCase.GetFloatIndex("Moment1") << ", " << incomingCase.GetFloatIndex("Moment2") << "," << incomingCase.GetFloatIndex("Moment3") << std::endl;

	if(mode < 20)
		ret = ReviseFirst<ImageType>(image, incomingCase, num, sol, cat);
	else
		ret = ReviseSecond<ImageType>(image, incomingCase, num, sol,cat);

	return ret;
}




template<class ImageType>
int ReviseFirst(typename ImageType::Pointer image, CBRCase &incomingCase, int num, typename ImageType::Pointer sol, std::string cat) {

	//WriteImageFile<ImageType>(image, cat+"revFirst-"+ToString(num)+".mhd");
	//WriteImageFile<ImageType>(createMatchImageToImage<ImageType>(image, sol), cat+"ovFirst-"+ToString(num)+".mhd");

	int errorNo = JudgeFirst<ImageType>(image,incomingCase, num);
	
	switch(errorNo) {
		case 2: 	// no ventricle found
			incomingCase.SetIntIndex("firstThresh", incomingCase.GetIntIndex("firstThresh") + 1); 
			break;

		case 3: 
			incomingCase.SetIntIndex("firstThresh", incomingCase.GetIntIndex("firstThresh") - 1);
			//incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("firstThresh") + 5); 
			break;
	
		case 4:
			incomingCase.SetIntIndex("firstThresh", incomingCase.GetIntIndex("firstThresh") - 1);
			// quit first revision.
			return 1;

		case 5:
			return 1;

		case 6:
			incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("firstThresh"));
			incomingCase.SetIntIndex("firstThresh", incomingCase.GetIntIndex("firstThresh") - 1);
	}

	return errorNo;
}


/*

	Return codes:
	2: no region segmented. Set higher threshold.
	3: Lower threshold
	4: Lower threshold, then quit.
	5: Keep this threshold and quit
	6: Lower threshold, set 2nd = 1st
	7: Skip this step. Use 0. 



*/
template<class ImageType>
int JudgeFirst(typename ImageType::Pointer image, CBRCase &incomingCase, int num){
	ImageType::IndexType centroid = GetCentroids<ImageType>(image);

	int vx = incomingCase.GetIntIndex("RegionX");
	int vy = incomingCase.GetIntIndex("RegionY");
	int vz = incomingCase.GetIntIndex("RegionZ");
	int x = incomingCase.GetIntIndex("FirstRegionX");
	int y = incomingCase.GetIntIndex("FirstRegionY");
	int z = incomingCase.GetIntIndex("FirstRegionZ");
	int mode = incomingCase.GetIntIndex("mode");
	std::cout << "Computed: " << centroid << " To remove: [" << x << ","<<y<<","<<z<<"]   Ventricle: [" << vx << ","<<vy<<","<<vz<<"]" << std::endl; 	
	float maxup = 2.0;
	float maxdn = 0.5;
	//tests..
	
	if(x==99 && y == 99 && z == 99) {
		return 5;
	}

	if(mode == 1){
		if(centroid[0] == 0 && centroid[1] == 0 && centroid[2] == 0) {
			//indicates no region found.	
			std::cout << "No region detected. Increasing first threshold" << std::endl;
			incomingCase.SetIntIndex("mode", 2);
			return 2;
		}

		if((abs(vx-centroid[0]) < 5) &&	(abs(vy-centroid[1]) < 5)&&	(abs(vz-centroid[2]) < 5)) {
			std::cout << "Ventricles detected, lowering first Thr." << std::endl;
			incomingCase.SetIntIndex("mode", 5);
			return 6;
		}
		if((abs(x-centroid[0]) < 5) &&	(abs(y-centroid[1]) < 5)&&	(abs(z-centroid[2]) < 5)) {
			std::cout << "Found to-be-removed at first try. Too much?." << std::endl;
			// test for size
			int regSize = GetSizeOfRegions<ImageType>(image)[0];
			if(regSize > (int)(incomingCase.GetIntIndex("size")*maxup)) {
				std::cout << "Region found is very big compared to size from case. Suspecting white-out" << std::endl;
				incomingCase.SetIntIndex("mode", 10);
				return 3;
			}
			else {
				std::cout << "Size looks fine, increasing. " << std::endl;
				incomingCase.SetIntIndex("mode", 3);
				return 2;
			}
		}		
	}
		
	// if the first region to pop up is the ventricle or the one to be removed
	if(mode == 2) {
		// region found where the ventricles should be
		if((abs(vx-centroid[0]) < 5) &&	(abs(vy-centroid[1]) < 5)&&	(abs(vz-centroid[2]) < 5)) {
			std::cout << "Increased and found ventricles, backing down and leaving it at that" << std::endl;
			incomingCase.SetIntIndex("mode", 19);
			return 4;
		}

		// region found is ok
		if((abs(x-centroid[0]) < 5) &&	(abs(y-centroid[1]) < 5)&&	(abs(z-centroid[2]) < 5)) {
			std::cout << "Increased and found region to be removed. Keep on increasing." << std::endl;
			incomingCase.SetIntIndex("mode", 3);
			return 2;
		}		

		std::cout << "Region found does not make sense. Keep on increasing too see what happens" << std::endl;
		return 2;
	}

	// mode 3 says we are working with the region to be found, and are increasing it.
	if(mode == 3) {
		if((abs(x-centroid[0]) <= 5) &&	(abs(y-centroid[1]) <= 5)&&	(abs(z-centroid[2]) <= 5)) {
			std::cout << "First region has nice centroid." << std::endl;
			int regSize = GetSizeOfRegions<ImageType>(image)[0];
			if(regSize > (int)(incomingCase.GetIntIndex("size")*maxup)) {
				std::cout << "Region is too big. Backing down and quiting." << std::endl;
				incomingCase.SetIntIndex("mode", 10);
				return 4;
			}
			else {
				std::cout << "Size looks fine, increasing. " << std::endl;
				incomingCase.SetIntIndex("mode", 3);
				return 2;
			}


			//return 2;
		}

		// did we just skip to the ventricles?
		if((abs(vx-centroid[0]) < 5) &&	(abs(vy-centroid[1]) < 5)&&	(abs(vz-centroid[2]) < 5)) {
			std::cout << "Looks like we skipped to the ventricles. Using last threshold" << std::endl;
			incomingCase.SetIntIndex("mode", 4);
			return 4;
		}

		if((abs(x-centroid[0]) > 5) ||	(abs(y-centroid[1]) > 5)||	(abs(z-centroid[2]) > 5)) {
			std::cout << "First region increased too much. Using last threshold" << std::endl;
			incomingCase.SetIntIndex("mode", 4);
			return 4;
		}		
	}


	// har funnet ventrikler, prver  senke til vi ikke gjr det lenger.
	if(mode == 5) {
		if((abs(vx-centroid[0]) < 5) &&	(abs(vy-centroid[1]) < 5)&&	(abs(vz-centroid[2]) < 5)) {
			std::cout << "Still finding ventricles" << std::endl;
			return 3;
		}

		if((abs(x-centroid[0]) <= 5) &&	(abs(y-centroid[1]) <= 5)&&	(abs(z-centroid[2]) <= 5)) {
			std::cout << "First region found. Will keep this threshold." << std::endl;
			return 5;
		}

	}

	if(mode == 10) {
		if((abs(vx-centroid[0]) < 5) &&	(abs(vy-centroid[1]) < 5)&&	(abs(vz-centroid[2]) < 5)) {
			std::cout << "Ventricles found from whiteout. Enter escape-ventricles mode." << std::endl;
			incomingCase.SetIntIndex("mode", 5);
			return 3;
		}

		if((abs(x-centroid[0]) <= 5) &&	(abs(y-centroid[1]) <= 5)&&	(abs(z-centroid[2]) <= 5)) {
			std::cout << "First region found. " << std::endl;
			int size = GetSizeOfRegions<ImageType>(image)[0];
			int max = (int)incomingCase.GetIntIndex("size")*maxup;
			if(size < max)  {
				std::cout << "Size is ok - will keep this threshold." << size << " < " << max << std::endl;
				return 5;
			}
			else {
				std::cout << "Size is too big, " << size << " >= " << max << std::endl; 
				return 3;
			}
		}

		std::cout << "Centroid still makes no sense. Keep lowering." << std::endl;
		return 3;
		
	}

	std::cout << "Ingen tester p centroid slo ut. Strrelse? " << std::endl;
	int regSize = GetSizeOfRegions<ImageType>(image)[0];
	if(regSize > (int)(incomingCase.GetIntIndex("size")*maxup)) {
		std::cout << "Region found is very big compared to size from case. Suspecting white-out" << std::endl;
		incomingCase.SetIntIndex("mode", 10);
		return 3;
	}
	else if(regSize < (int)(incomingCase.GetIntIndex("size")*maxdn)) {
		std::cout << "Region found is very small compared to size from case. Increasing (whithout just cause)." << std::endl;
		incomingCase.SetIntIndex("mode", 2); 
		return 2;		
	}
	else{
		std::cout << "Size is whitin acceptable bounds " << regSize << " =< " << (int)(incomingCase.GetIntIndex("size")*maxup) <<std::endl;
	}



	return 1;
}


template<class ImageType>
int ReviseSecond(typename ImageType::Pointer image, CBRCase &incomingCase, int num, typename ImageType::Pointer sol, std::string cat) {
	//WriteImageFile<ImageType>(image, cat+"revSec-"+ToString(num)+".mhd");
	//WriteImageFile<ImageType>(createMatchImageToImage<ImageType>(image, sol), cat+"ovFirst-"+ToString(num)+".mhd");

	int errorNo = JudgeSecond<ImageType>(image,incomingCase, num);

	switch(errorNo) {
		//default:
		//	return 1;

		case 21:
			// found to be too large.
			incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("secondThresh") -1 );
			return 21;
			break;
		
		case 23:
			//found to be too small
			incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("secondThresh") + 1);
			return 22;
			break;
		
		case 24: 
			incomingCase.SetIntIndex("secondThresh", incomingCase.GetIntIndex("secondThresh") - 1);
			break;
	
		case 26:
			return 26;
	}
	return 1;
}

template<class ImageType>
int JudgeSecond(typename ImageType::Pointer image, CBRCase &incomingCase, int num) {
	ImageType::IndexType centroid = GetCentroids<ImageType>(image);

	int vx = incomingCase.GetIntIndex("RegionX");
	int vy = incomingCase.GetIntIndex("RegionY");
	int vz = incomingCase.GetIntIndex("RegionZ");
	int x = incomingCase.GetIntIndex("FirstRegionX");
	int y = incomingCase.GetIntIndex("FirstRegionY");
	int z = incomingCase.GetIntIndex("FirstRegionZ");
	int mode = incomingCase.GetIntIndex("mode");
	int wantedSize = incomingCase.GetIntIndex("size");
	int size = GetSizeOfRegions<ImageType>(image)[0];
	int max = (int)wantedSize*2;
	int min = (int)wantedSize*0.5;
	float maxmom = 2;
	std::cout << "Computed: " << centroid << "Ventricle: [" << vx << ","<<vy<<","<<vz<<"] (To remove: [" << x << ","<<y<<","<<z<<"])"    << std::endl; 	

		// everything is correct, increase or decrease based on inplaceVar alone.
	if(mode == 26) {
		//float variance = incomingCase.GetFloatIndex("InPlaceVarianceTmp");
		//float CorVariance = incomingCase.GetFloatIndex("InPlaceVariance");
		//if(variance < (int)(CorVariance*0.8)) {
		//	std::cout << "Too small variance. Increasing." << std::endl;
		//	if(incomingCase.GetStringIndex("Direction") == "down" ) {std::cout << "Loop detected. Quit." << std::endl; return 1;}
		//	incomingCase.SetStringIndex("Direction", "up");
		//	return 23;
		//}
		//if(variance > (int)(CorVariance*1.2)) {
		//	std::cout << "Variance too big. Decreasing." << std::endl;
		//	if(incomingCase.GetStringIndex("Direction") == "up" ) {std::cout << "Loop detected. Quit." << std::endl; return 1;}
		//	incomingCase.SetStringIndex("Direction", "down");
		//	return 21;
		//}
		//std::cout << "Variance: My :" << variance << " vs from case: " <<  CorVariance <<std::endl;
		return 1;
	}

	// half ventricle found- mode.
	if(mode == 25) {
		if(size > max) {
			std::cout << "Region is way to large. Quit. " << std::endl;
			return 24;
		}
		if((abs(vx-centroid[0]) < 5) && (abs(vy-centroid[1]) < 5) && (abs(vz-centroid[2]) < 5) ) {
			std::cout << "From half-ventricle-mode, acceptable. Quitting." << std::endl;
			return 1;
		}
		
	}

	//is size way to big?
	if(size > (max)) {
		std::cout << "Big leak suspected detected by size. Lowering." << std::endl;
		if(incomingCase.GetIntIndex("tooSmall") != -1) { 
			std::cout << "Was too small, now too big. Qutting in frustration " << std::endl;
			return 1;
		}
		incomingCase.SetIntIndex("tooBig", 1);
		incomingCase.SetIntIndex("mode", 21);
		return 21;
	}
	else if(size < min) {
		std::cout << "Size is too small" << std::endl;
		if(incomingCase.GetIntIndex("tooBig") != -1) { 
			std::cout << "Was too big, now too small. Qutting in frustration " << std::endl;
//			incomingCase.SetIntIndex("mode", 26);
			return 1;
		}
		// found half ventricle?
		if((abs(centroid[1] - vy) < 5) && ((abs(centroid[2] - vz) < 5))) {
			std::cout << "Suspecting halv ventricle found. Increasing." << std::endl;
			incomingCase.SetIntIndex("mode", 25);
			return 23;
		}
		incomingCase.SetIntIndex("tooSmall" , 1);
		incomingCase.SetIntIndex("mode", 23);
		return 23;
	}
	else {
		std::cout << "Size ok, " << size << " <= " << max << std::endl;
	}


	std::vector<float> moments = GetMoments<ImageType>(image);
	if(moments[2] > (int)(incomingCase.GetFloatIndex("Moment3")*maxmom)) {
		std::cout << "Leak detected by Moment 3. Lowering" << std::endl;
		incomingCase.SetIntIndex("mode", 22);
		return 21;
	}
	else {
		std::cout << "Moment ok, " << moments[2] << " <= " << (int)(incomingCase.GetFloatIndex("Moment3")*maxmom) << std::endl;		
	}

	
	if(mode == 21) {
		std::cout << "Reduce leak-mode" << std::endl;
		if((abs(vx-centroid[0]) < 5) && (abs(vy-centroid[1]) < 5) && (abs(vz-centroid[2]) < 5) ) {
			std::cout << "Centroid approved. " << std::endl;
			if(size > (int)(wantedSize*2)) {
				std::cout << "Size is too large, lowering. " << size << " vs " << wantedSize << std::endl;
				return 21;
			}
			else {
				std::cout << "Size seems ok " << size << " =< " << (wantedSize*2) <<std::endl;
				if(moments[2] > (incomingCase.GetFloatIndex("Moment3")*maxmom)) {
					std::cout << "Failed on Moment3" << moments[2] << " > " << (incomingCase.GetFloatIndex("Moment3")*maxmom) << std::endl;
//					std::cout << "moments[2]
					//					incomingCase.set
					return 21;
				}
				else {
					std::cout << "Moment3 ok" << moments[2] << " <= " << (incomingCase.GetFloatIndex("Moment3")*maxmom) << std::endl;
				}
			}
			return 1;
			//incomingCase.SetIntIndex("mode", 26);
			//return 26;
		}
		else {
			// keep reducing.
			return 21;
		}
	}

	if(mode == 22) {
		if(abs(moments[3] > incomingCase.GetFloatIndex("Moment3")*maxmom)) {
			std::cout << "Still failing on Moment3" << std::endl;
			return 21;
		}
	}

	if(mode == 23) {
		std::cout << "Size increasing-mode" << std::endl;
		int sizeNow = GetSizeOfRegions<ImageType>(image)[0];
		if(sizeNow > max) {
			std::cout << "Size is now too high. Qutting in despair. " << std::endl;
			return 1;
		}
		else {
			if(moments[2] > (incomingCase.GetFloatIndex("Moment3")*2.5)) {
				std::cout << "Size is ok - but moment 3 too high. Qutting." << moments[2] << " > " << (incomingCase.GetFloatIndex("Moment3")*2.5) << std::endl;
				return 1; //-- IKKE MED??
			}
		}
		// keep on increasing.
		return 23;
	}


	//if(mode != 26) {
	//	incomingCase.SetIntIndex("mode" , 26);
	//	return 26;
	//}

	std::cout << "JudgeSecond returnerer uten  ha sltt ut p noen tester. Slutter." << endl;
	return 1;
}

#endif CBR_I_F
