#ifndef __CBRITICH_
#define __CBRITICH_

#include <itkImage.h>
#include <itkImageRegion.h>
#include <itkRegionOfInterestImageFilter.h>
#include <math.h>

template<class T>
T MIN(T a, T b) {
	if(a <= b) return a;
	return b;
}

template<class T>
T MAX(T a, T b) {
	if(a >= b) return a;
	return b;
}


template<class T>
float fpi(int intensity, typename T::Pointer window) {
//	std::cout << "start" << std::flush << std::endl;
	//iterate through window and find city distance

	//the coordinates of the pixel with intensity is the same as the centre of Window
	// the coordinates from the full image are lost/not available/not important
	T::SizeType size = GetSizeOfImage<T>(window);
	int x = (size[0]-1)/2;
	int y = (size[1]-1)/2;
	int z = (size[2]-1)/2;

	//best so far and other variables :)
	float min=100000;
	float tmp=0;
	T::IndexType index;

	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;
	ConstIteratorType it(window, window->GetRequestedRegion());

//	std::cout << "beginning fpi iteration" << std::flush << std::endl;

	it.Begin();
	while(!it.IsAtEnd()) {
		index = it.GetIndex();
		tmp = abs(index[0]-x) + abs(index[1]-y) + abs(index[2]-z) + abs(intensity-it.Get());
//		std::cout << "fpi : " << index << std::endl <<std::flush ;

		if(tmp < min)
			min = tmp;
	
		++it;
	}	

	return min;
}
			



template<class T>
float Hausdorff(typename T::Pointer newImage, typename T::Pointer caseImage, int size) {

	//size must be odd
	if(!(size & 1)) size++;

	int x,y,z;
	int offset = (size-1)/2;
	x = y = z = offset;

	// get size of newImage and caseImage, only use shared volume
	T::SizeType searchSize = getSearchWindowSize<T>(newImage, caseImage); 	
	T::IndexType centerIndex;
	T::Pointer subNew;
	T::Pointer subCase;
	float exp = 2;
	float dab = 0;
	float dba = 0;
	float sum = 0;

	std::cout << searchSize << std::flush << std::endl;

	for(x=offset; x < searchSize[0]-offset; x++) {
		for(y=offset; y < searchSize[1]-offset; y++) {
			for(z=offset; z< searchSize[2]-offset; z++) {
				subNew = GetSubRegionInImage<T>(newImage, size, x, y, z);
				subCase = GetSubRegionInImage<T>(caseImage, size, x, y, z);
			
				// get distance from pixel to image
				centerIndex[0] = x, centerIndex[1] = y, centerIndex[2] = z;
//				std::cout << "calling : " << x << "," << y <<  "," << z << std::endl <<std::flush ;
				dab = fpi<T>(newImage->GetPixel(centerIndex), subCase);
				dba = fpi<T>(caseImage->GetPixel(centerIndex), subNew);
				sum += pow((dab + dba), exp);
//				sum = 0.0;	
			}
		}
//		std::cout << "x: " << x << std::endl;
	}
	sum = pow(sum, (1/exp));
	sum /= (searchSize[0] * searchSize[1] * searchSize[2]);
	sum /= 2;
	//num colors. probably not 35000.
	sum /= 35000;

	std::cout << "Hausdorff: " << sum << std::endl;

	return sum/2;
}





/*
	Not used. moved to CBRImageFunctions.h
	will clean up later.
*/

template<class T>
typename T::Pointer getSubVolumeDepr(typename T::Pointer image, typename T::IndexType min, typename T::IndexType max) {
	/*
		This returns the region from an image specified by coords min and max
	*/

	T::RegionType region;

	typedef itk::RegionOfInterestImageFilter<T,T> FilterType;
	FilterType::Pointer filter = FilterType::New();

	//compute size in each direction
	T::SizeType size;
	size[0] = max[0] - min[0];
	size[1] = max[1] - min[1];
	size[2] = max[2] - min[2];

	region.SetSize(size);
	region.SetIndex(min);

	//set up filter with this region of interest
	filter->SetRegionOfInterest(region);
	filter->SetInput(image);

	T::Pointer retImage = filter->GetOutput();

	// update image to process pipeline and get result.
	try{
		retImage->Update();
	}
	catch(itk::ExceptionObject &exp) {
		std::cerr << exp;
	}

	return retImage;
}



template<class T>
int GetIntersection(typename T::IndexType amin, typename T::IndexType amax, 
					 typename T::IndexType bmin, typename T::IndexType bmax,
					 typename T::IndexType &minRet, typename T::IndexType &maxRet) {
	/*
		To get the subregion take the largest of the mins and smallest of the maxs
		for each dimension to get overlap.
		Check that they actually overlap.
	*/

	 // 0 = ok
	int ireturn = 0;

	minRet[0] = MAX<int>(amin[0], bmin[0]);
	minRet[1] = MAX<int>(amin[1], bmin[1]);
	minRet[2] = MAX<int>(amin[2], bmin[2]);
	maxRet[0] = MIN<int>(amax[0], bmax[0]);
	maxRet[1] = MIN<int>(amax[1], bmax[1]);
	maxRet[2] = MIN<int>(amax[2], bmax[2]);

	// check for actual overlap
	if(minRet[0] > maxRet[0]) ireturn= -1;
	if(minRet[1] > maxRet[1]) ireturn= -1;
	if(minRet[2] > maxRet[2]) ireturn= -1;
					 
	return ireturn;
}
template<class T>
typename T::Pointer GetSubRegionInImage(typename T::Pointer image, int size, int x, int y, int z) {

	T::RegionType region;

	

	typedef itk::RegionOfInterestImageFilter<T,T> FilterType;
	FilterType::Pointer filter = FilterType::New();

	//size is odd. if 9, have 4 on each side.
	if(!(size & 1)) size++;
	int offset = size/2;


	//determine borders for requested region
	T::SizeType maxSize = GetSizeOfImage<T>(image);
	
//	std::cout << offset << (x-offset) << std::endl;

	if(x-offset < 0) x=offset;
	if(y-offset < 0) y=offset;
	if(z-offset < 0) z=offset;

	if((x + offset) > maxSize[0]) x = maxSize[0] -offset;
	if((y + offset) > maxSize[1]) y = maxSize[1] -offset;
	if((z + offset) > maxSize[2]) z = maxSize[2] -offset;


	// define size and starting point of region
	T::SizeType Rsize;
	T::IndexType index;
	index[0] = x-offset;
	index[1] = y-offset;
	index[2] = z-offset;

	Rsize[0] = size;
	Rsize[1] = size;
	Rsize[2] = size;
	region.SetIndex(index);
	region.SetSize(Rsize);
	
	// set up pipe
	filter->SetRegionOfInterest(region);
	filter->SetInput(image);

	// set up return image
	T::Pointer retImage = filter->GetOutput();

	try{
		retImage->Update();
	}
	catch(itk::ExceptionObject & exep) {
		std::cerr << exep << "Exception caught";
		std::cout << offset << (x-offset) << std::endl;
	}

	return retImage;
}



template<class T>
typename T::SizeType getSearchWindowSize(typename T::Pointer a, typename T::Pointer b) {
		
	T::SizeType asize = GetSizeOfImage<T>(a);
	T::SizeType bsize = GetSizeOfImage<T>(b);

	T::SizeType retSize;
	retSize[0] = MIN<int>(asize[0], bsize[0]);
	retSize[1] = MIN<int>(asize[1], bsize[1]);
	retSize[2] = MIN<int>(asize[2], bsize[2]);

	return retSize;

}



template<class T>
void getBoundaryOfHeadInImage(typename T::Pointer image, typename T::IndexType &minRet, typename T::IndexType &maxRet) {
/*
	This is used to get the bounding box of the head in the volume.
	this is further used to assess similarity at a very coarse level.
*/

	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;


	// get an image with (hopefully) only interesting borders
	T::Pointer thrImage = markHeadRegionGrowing<T>(image);
	
	//set up iterator
	thrImage->Update();
	ConstIteratorType it(thrImage, thrImage->GetRequestedRegion());

	// variables to hold maxima
	int xmin = std::numeric_limits<int>::max();
	int xmax = 0;
	int ymin = std::numeric_limits<int>::max();
	int ymax = 0;
	int zmin = std::numeric_limits<int>::max();
	int zmax = 0;

	// get coordinates of non-zero pixel

	T::IndexType index;
	//go through iterator and get top, bottom,left,right,front and rear maxima	
	it.GoToBegin();
	while(!it.IsAtEnd()) {
		if(it.Get() > 20000) {
			index = it.GetIndex();
			// x
			if(index[0] < xmin) xmin = index[0];
			if(index[0] > xmax) xmax = index[0];
			// y
			if(index[1] < ymin) ymin = index[1];
			if(index[1] > ymax) ymax = index[1];
			// z
			if(index[2] < zmin) zmin = index[2];
			if(index[2] > zmax) zmax = index[2];

		}
		++it;
	}
	
	minRet[0] = xmin;
	minRet[1] = ymin;
	minRet[2] = zmin;

	maxRet[0] = xmax;
	maxRet[1] = ymax;
	maxRet[2] = zmax;

}



template<class T>
typename T::Pointer markHeadRegionGrowing(typename T::Pointer image) {
/*
	
	Some of the volumes have some "bad" values besides the head, residue 
	that makes it impossible to find the head looking only for x > xmax 
	whitin a threshold.

*/
	typedef itk::CurvatureFlowImageFilter<T,T> SmoothingFilterType;
	typedef itk::NeighborhoodConnectedImageFilter<T,T> NeighborFilterType;

	SmoothingFilterType::Pointer smoothing = SmoothingFilterType::New();
	NeighborFilterType::Pointer neigh = NeighborFilterType::New();

	smoothing->SetInput(image);
	neigh->SetInput(smoothing->GetOutput());

	//set up seed point, a little left of center to hit white matter
	T::IndexType index;
	index[0] = 100;
	index[1] = 30;
	index[2] = 137;
	
	T::SizeType radius;
	radius[0] = 0;
	radius[1] = 0;
	radius[2] = 0;

	neigh->SetSeed(index);
	neigh->SetRadius(radius);

	//set up some boundaries to fill head.
	neigh->SetLower(5000);
	neigh->SetUpper(40000);
	neigh->SetReplaceValue(32000);

	try{
		neigh->Update();
	}
	catch(itk::ExceptionObject &  exep)
	{
		std::cerr << exep << "Exception caught";
	}

	return neigh->GetOutput();
}



template<class T>
float HausdorffSubVolumes(typename T::Pointer a, typename T::Pointer b, int size) {

/*
	Method to take to images to be compared.
	Each image will be processed to find the subvolume containing the head.
	After this, the intersection of these volumes will be computed.
	Finally, Hausdorff is called on the two intersections.

*/

	// boundary points in the images
	T::IndexType amin;
	T::IndexType amax;
	T::IndexType bmin;
	T::IndexType bmax;

	// get boundary, call by reference
	getBoundaryOfHeadInImage<T>(a, amin, amax);
	getBoundaryOfHeadInImage<T>(b, bmin, bmax);

	// debug
	//std::cout << amin << amax << std::endl;
	//std::cout << bmin << bmax << std::endl;

	//intersection of the two volumes
	T::IndexType volMin;
	T::IndexType volMax;
	
	int check = GetIntersection<T>(amin, amax, bmin, bmax, volMin, volMax);

	//no overlap
	if(check == -1)
		return 0.0;

//	std::cout << volMin << volMax << std::endl;

	T::Pointer smalla = getSubVolume<T>(a, volMin, volMax);
	T::Pointer smallb = getSubVolume<T>(b, volMin, volMax);


	return Hausdorff<T>(smalla, smallb, size);
}

#endif
