/*=========================================================================

Program:   Insight Segmentation & Registration Toolkit
Module:    $RCSfile: ThresholdSegmentationLevelSetImageFilter.cxx,v $
Language:  C++
Date:      $Date: 2005/11/19 16:31:51 $
Version:   $Revision: 1.35 $

Copyright (c) Insight Software Consortium. All rights reserved.
See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

This software is distributed WITHOUT ANY WARRANTY; without even 
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#if defined(_MSC_VER)
#pragma warning ( disable : 4786 )
#endif

#ifdef __BORLANDC__
#define ITK_LEAN_AND_MEAN
#endif





#include "itkImage.h"
// Software Guide : BeginCodeSnippet
#include "itkThresholdSegmentationLevelSetImageFilter.h"
// Software Guide : EndCodeSnippet
#include "itkFastMarchingImageFilter.h"
#include "itkBinaryThresholdImageFilter.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkZeroCrossingImageFilter.h"

template<class ImageType>
typename ImageType::Pointer LevelSet(typename ImageType::Pointer input, typename ImageType::Pointer speed, typename ImageType::IndexType InSeed, float minThr, float maxThr, float initDist) {
	//int LevelSet( int argc, char *argv[] )
	//{
	//if( argc < 8 )
	//  {
	//  std::cerr << "Missing Parameters " << std::endl;
	//  std::cerr << "Usage: " << argv[0];
	//  std::cerr << " inputImage  outputImage";
	//  std::cerr << " seedX seedY InitialDistance";
	//  std::cerr << " LowerThreshold";
	//  std::cerr << " UpperThreshold";
	//  std::cerr << " [CurvatureScaling == 1.0]";
	//  std::cerr << std::endl;
	//  return 1;
	//  }

	//  Software Guide : BeginLatex
	//  
	//  We define the image type using a particular pixel type and
	//  dimension. In this case we will use 2D \code{float} images.
	//
	//  Software Guide : EndLatex 

	// Software Guide : BeginCodeSnippet
	typedef   float           InternalPixelType;
	const     unsigned int    Dimension = 3;
	typedef itk::Image< InternalPixelType, Dimension >  InternalImageType;
	// Software Guide : EndCodeSnippet

	typedef float OutputPixelType;
	typedef itk::Image< OutputPixelType, Dimension > OutputImageType;
	typedef itk::BinaryThresholdImageFilter<InternalImageType, OutputImageType>
		ThresholdingFilterType;

	ThresholdingFilterType::Pointer thresholder = ThresholdingFilterType::New();

	thresholder->SetLowerThreshold( -1000.0 );
	thresholder->SetUpperThreshold(     0.0 );

	thresholder->SetOutsideValue(  0  );
	thresholder->SetInsideValue(  255 );

	typedef  itk::ImageFileReader< InternalImageType > ReaderType;
	typedef  itk::ImageFileWriter<  OutputImageType  > WriterType;

	ReaderType::Pointer reader = ReaderType::New();
	WriterType::Pointer writer = WriterType::New();

	//  reader->SetFileName( argv[1] );
	//  writer->SetFileName( argv[2] );


	//  We now declare the type of the \doxygen{FastMarchingImageFilter} that
	//  will be used to generate the initial level set in the form of a distance
	//  map.
	//
	typedef  itk::FastMarchingImageFilter< InternalImageType, InternalImageType >
		FastMarchingFilterType;

	FastMarchingFilterType::Pointer  fastMarching = FastMarchingFilterType::New();

	//  Software Guide : BeginLatex
	//  
	//  The following lines instantiate a
	//  ThresholdSegmentationLevelSetImageFilter using the \code{New()} method.
	//
	//  Software Guide : EndLatex 

	// Software Guide : BeginCodeSnippet
	typedef  itk::ThresholdSegmentationLevelSetImageFilter< InternalImageType, 
		InternalImageType > ThresholdSegmentationLevelSetImageFilterType;
	ThresholdSegmentationLevelSetImageFilterType::Pointer thresholdSegmentation =
		ThresholdSegmentationLevelSetImageFilterType::New();
	// Software Guide : EndCodeSnippet



	//  Software Guide : BeginCodeSnippet
	thresholdSegmentation->SetPropagationScaling( 1.0 );
	//if ( argc > 8 )
	//  {
	//  thresholdSegmentation->SetCurvatureScaling( atof(argv[8]) );
	//  }
	//else
	//{
		thresholdSegmentation->SetCurvatureScaling( 1.0 );
		//}

		//  Software Guide : EndCodeSnippet 

		//  The level set solver will stop if the convergence criteria has been
		//  reached or if the maximum number of iterations has elasped.  The
		//  convergence criteria is defined in terms of the root mean squared (RMS)
		//  change in the level set function. When RMS change for an iteration is
		//  below a user-specified threshold, the solution is considered to have
		//  converged.
		thresholdSegmentation->SetMaximumRMSError( 0.02 );
		thresholdSegmentation->SetNumberOfIterations( 1200 );

		//    thresholdSegmentation->SetMaximumRMSError( atof(argv[8]) );
		//    thresholdSegmentation->SetNumberOfIterations( atoi(argv[9]) );


		// Software Guide : BeginCodeSnippet
		thresholdSegmentation->SetUpperThreshold( maxThr );
		thresholdSegmentation->SetLowerThreshold( minThr );
		thresholdSegmentation->SetIsoSurfaceValue(0.0);
		// Software Guide : EndCodeSnippet

		//  Software Guide : BeginLatex
		//  
		//  The filters are now connected in a pipeline indicated in
		//  Figure~\ref{fig:ThresholdSegmentationLevelSetImageFilterDiagram}.
		//  Remember that before calling \code{Update()} on the file writer object,
		//  the fast marching filter must be initialized with the seed points and
		//  the output from the reader object.  See previous examples and the
		//  source code for this section for details.
		//
		//  Software Guide : EndLatex 

		// Software Guide : BeginCodeSnippet
//		thresholdSegmentation->SetInput( fastMarching->GetOutput() );
		thresholdSegmentation->SetInput( speed );
		thresholdSegmentation->SetFeatureImage(input );
		thresholder->SetInput( thresholdSegmentation->GetOutput() );
		//  writer->SetInput( thresholder->GetOutput() );
		// Software Guide : EndCodeSnippet
		ImageType::Pointer returnImage = thresholder->GetOutput();
		//
		//  The FastMarchingImageFilter requires the user to provide a seed
		//  point from which the level set will be generated. The user can actually
		//  pass not only one seed point but a set of them. Note the the
		//  FastMarchingImageFilter is used here only as a helper in the
		//  determination of an initial Level Set. We could have used the
		//  \doxygen{DanielssonDistanceMapImageFilter} in the same way.
		//
		//  The seeds are passed stored in a container. The type of this
		//  container is defined as \code{NodeContainer} among the
		//  FastMarchingImageFilter traits.
		//
		typedef FastMarchingFilterType::NodeContainer           NodeContainer;
		typedef FastMarchingFilterType::NodeType                NodeType;

		NodeContainer::Pointer seeds = NodeContainer::New();

		InternalImageType::IndexType  seedPosition;

		seedPosition[0] = 20;  //atoi( argv[3] );
		seedPosition[1] = 20;  //atoi( argv[4] );
		seedPosition[2] = 20;
		//  Nodes are created as stack variables and initialized with a value and an
		//  \doxygen{Index} position. Note that here we assign the value of minus the
		//  user-provided distance to the unique node of the seeds passed to the
		//  FastMarchingImageFilter. In this way, the value will increment
		//  as the front is propagated, until it reaches the zero value corresponding
		//  to the contour. After this, the front will continue propagating until it
		//  fills up the entire image. The initial distance is taken here from the
		//  command line arguments. The rule of thumb for the user is to select this
		//  value as the distance from the seed points at which he want the initial
		//  contour to be.

		const double initialDistance = initDist; 

		NodeType node;
		NodeType node2;

		const double seedValue = - initialDistance;

		node.SetValue( seedValue );
		node.SetIndex( seedPosition );

		ImageType::IndexType index2;
		index2[0] = 37; 
		index2[1] = 13; 
		index2[2] = 23;
		node2.SetValue( seedValue );
		node2.SetIndex( index2 );
		//
		//  The list of nodes is initialized and then every node is inserted using
		//  the \code{InsertElement()}.
		seeds->Initialize();
		seeds->InsertElement( 0, node );
		seeds->InsertElement( 1, node2 );

		//  The set of seed nodes is passed now to the
		//  FastMarchingImageFilter with the method
		//  \code{SetTrialPoints()}.
		fastMarching->SetTrialPoints(  seeds  );

		//  
		//  Since the FastMarchingImageFilter is used here just as a
		//  Distance Map generator. It does not require a speed image as input.
		//  Instead the constant value $1.0$ is passed using the
		//  \code{SetSpeedConstant()} method.

		fastMarching->SetSpeedConstant( 1.0 );



		// Software Guide : BeginCodeSnippet
		try
		{
			input->Update();
			fastMarching->SetOutputSize( 
				input->GetBufferedRegion().GetSize() );

			returnImage->Update();

		}
		catch( itk::ExceptionObject & excep )
		{
			std::cerr << "Exception caught !" << std::endl;
			std::cerr << excep << std::endl;
		}
		// Software Guide : EndCodeSnippet

		// Print out some useful information 
		std::cout << std::endl;
		std::cout << "Max. no. iterations: " << thresholdSegmentation->GetNumberOfIterations() << std::endl;
		std::cout << "Max. RMS error: " << thresholdSegmentation->GetMaximumRMSError() << std::endl;
		std::cout << std::endl;
		std::cout << "No. elpased iterations: " << thresholdSegmentation->GetElapsedIterations() << std::endl;
		std::cout << "RMS change: " << thresholdSegmentation->GetRMSChange() << std::endl;


		// We write out some intermediate images for debugging.  These images can
		// help tune parameters.
		//
		typedef itk::ImageFileWriter< InternalImageType > InternalWriterType;

		InternalWriterType::Pointer mapWriter = InternalWriterType::New();
		mapWriter->SetInput( fastMarching->GetOutput() );
		mapWriter->SetFileName("fastMarchingImage.mhd");
		mapWriter->Update();

		InternalWriterType::Pointer speedWriter = InternalWriterType::New();
		speedWriter->SetInput( thresholdSegmentation->GetSpeedImage() );
		speedWriter->SetFileName("speedTermImage.mhd");
		speedWriter->Update();

		return returnImage;
	}



