#ifndef CBRSTAT_H
#define CBRSTAT_H

#include <itkScalarImageToHistogramGenerator.h>
#include <itkImage.h>
#include <itkConnectedComponentImageFilter.h>
#include <itkRelabelComponentImageFilter.h>

#include <Math.h>
#include <Map>
#include <Vector>
#include <Utility>
#include <Set>


/***************************************************************************

	Computes and returns the histogram of the given image. 
	This was as clean as i could get it.

***************************************************************************/
template<class T, class HistogramGeneratorType>
const typename HistogramGeneratorType::Pointer GetHistogramFromImage(typename T::Pointer image, int numBins, float marginalScale) {
	
	typedef HistogramGeneratorType::HistogramType histType;

	HistogramGeneratorType::Pointer histGen = HistogramGeneratorType::New();

	histGen->SetInput(image);

	histGen->SetNumberOfBins(numBins);
	histGen->SetMarginalScale(marginalScale);
	histGen->SetHistogramMin(  0 );
	histGen->SetHistogramMax( 256 );


	histGen->Compute();

	const histType * histogram = histGen->GetOutput();

//	SmoothHistogram(histogram, numBins);

//	WriteHistogramToDisk<histType>(histogram);	

	return histGen;
}





template<class ImageType> 
std::map<int,int> getHistogramOfSubImage(typename ImageType::Pointer image, typename ImageType::Pointer sol, int binSize, std::string mode = "other") {
	std::map<int, int> histogram;
		
		// set up iterators
	typedef itk::ImageRegionConstIterator<ImageType> ConstIteratorType;
	ConstIteratorType solIt(sol, sol->GetRequestedRegion());
	ConstIteratorType origIt(image, image->GetRequestedRegion());

	solIt.GoToBegin();
	origIt.GoToBegin();
	while(!origIt.IsAtEnd()) {
		if(mode == "Ventricle") {
			if(solIt.Get() == VentricleValue){
				int bin = (int)(origIt.Get() / binSize);
				histogram[bin]++;
			}
		}
		else {
			if(solIt.Get() != VentricleValue){
				int bin = (int)(origIt.Get() / binSize);
				histogram[bin]++;
			}
		}
		++solIt;
		++origIt;
	}

	//std::map<int,int>::const_iterator it;
	//for(it = histogram.begin() ; it != histogram.end(); it++) {
	//	std::cout << it->first << std::endl;	
	//}
	//std::cout << "-------------------------------------"<<std::endl;
	//for(it = histogram.begin() ; it != histogram.end(); it++) {
	//	std::cout << it->second << std::endl;	
	//}

	int max = (int)(256 / binSize);
	for(unsigned int i = 0; i<max; i++)
		std::cout << histogram[i] << std::endl;


	return histogram;
}





template<class ImageType>
int getVolumeOfLargestSegment(typename ImageType::Pointer image){

	typedef itk::ConnectedComponentImageFilter<ImageType, ImageType> ConnectType;
	typedef itk::RelabelComponentImageFilter<ImageType, ImageType> RelabelType;
	ConnectType::Pointer connector = ConnectType::New();
	RelabelType::Pointer relabler = RelabelType::New();

	connector->SetInput(image);
	relabler->SetInput(connector->GetOutput());
	ImageType::Pointer labeled = relabler->GetOutput();

	try{
		labeled->Update();
	}
	catch(itk::ExceptionObject &exp) {
		std::cerr << exp;
	}

	
	typedef itk::ImageRegionConstIterator<ImageType> itType;
	itType it(labeled, labeled->GetRequestedRegion());
	it.GoToBegin();
	typedef std::pair<ImageType::IndexType, ImageType::IndexType> pairType;
	pairType minmax; 
	ImageType::IndexType min;
	ImageType::IndexType max;
	std::vector<int> visited;
	unsigned int size = 0;
	unsigned int maxS = 0;
	while(!it.IsAtEnd()) {
		if(it.Get() != 0 && notContains(it.Get(), visited)) {
//			std::cout << it.Get() << std::endl;
			minmax = getVentricleBoundary<ImageType>(labeled, it.Get());		
			min = minmax.first;
			max = minmax.second;
			size = (max[0]-min[0])*(max[1]-min[1])*(max[2]-min[2]);
			visited.push_back(it.Get());
			if(size > maxS) maxS = size;
		}
		++it;
	}

	std::cout << "Max: " << maxS << std::endl;
//	std::cout << relabler->GetNumberOfObjects() << std::endl;

	//const std::vector<unsigned int> sizes;
	//if(relabler->GetNumberOfObjects() > 0)
	//return relabler->GetSizeOfObjectsInPixels()[0];
	return maxS;
//	return sizes[0];
}


/*********************************************************************************************************

scan through, set each segment to a unique color, starting with 1

*********************************************************************************************************/
int notContains(int tall, std::vector<int> vector) {
	for(int i = 0; i < vector.size(); i++) {
		if(vector.at(i) == tall)
			return 0;
	}	
	return 1;
}


/*********************************************************************************************************
*********************************************************************************************************/
template<class HistogramType>
void SmoothHistogram(HistogramType *histogram, int numBins) {

	// get instance id
//	HistogramType::InstanceIdentifier id = histogram->GetInstanceIdentifier();

	// get an index..
	HistogramType::IndexType index;
	

	for(unsigned int i = 1; i < (numBins-1) ; i++) {
//		index = histogram->GetIndex(id);
	}
}





/******************************************************************************************

	Computes H(g) = N(g) / S = histgram(color) / Number of pixels. (probability of color)

******************************************************************************************/
template<class HistogramType>
std::vector<float> NormalizeHistogram(const HistogramType * histogram) {

	// Total number of pixels
//	float totalPix = GetNumberOfPixels<HistogramType>(histogram);
	HistogramType::FrequencyType freq = histogram->GetTotalFrequency();
	//std::cout << "TotalFreq: " << freq << std::endl;

	// iterators over histogram.
	HistogramType::ConstIterator it = histogram->Begin();
	HistogramType::ConstIterator end = histogram->End();
	// vector to populate. Will be the histogram in a simpler packaging
	std::vector<float> retVector;
	
	while(it != end) {
		retVector.push_back((float)(it.GetFrequency()) / freq);
		++it;
	}

	// return normalized vector.
	return retVector;
}



/******************************************************************************

		Entry point for calculating the statistics on histogram

******************************************************************************/
template<class ImageType>
std::vector<float> GetHistogramStatistics(typename ImageType::Pointer image) {

	std::vector<float> retVector;

	// type of histogram we will be working with.
	typedef itk::Statistics::ScalarImageToHistogramGenerator<ImageType> histTypeGenerator;
	typedef histTypeGenerator::HistogramType HistogramType;

	// get the actual histogram.
	histTypeGenerator::Pointer ht = GetHistogramFromImage<ImageType, histTypeGenerator>(image, 256, 10);
	const HistogramType *histogram = ht->GetOutput();

	// get histogram in normalized vector<float> form.
	std::vector<float> normHist = NormalizeHistogram<HistogramType>(histogram);

	// get values and save to vector
	float average = GetHistogramAverage(normHist);
	float var = GetHistogramVariance(normHist, average);
	float skew = GetHistogramSkewness(normHist, average, var);
	float kurtosis = GetHistogramKurtosis(normHist, average, var);
	float varcoeff = GetVarianceCoeff(var, average);
	float entropy = GetEntropy(normHist);
	std::vector<float> index = GetCentroidsColor<ImageType>(image, average);

	retVector.push_back(average);
	retVector.push_back(var);
	retVector.push_back(skew);
	retVector.push_back(kurtosis);
	retVector.push_back(varcoeff);
	retVector.push_back(entropy);
	retVector.push_back(index[0]);
	retVector.push_back(index[1]);
	retVector.push_back(index[2]);

	return retVector;
}


/**************************************************************************************************

	Returns the average color

**************************************************************************************************/
float GetHistogramAverage(std::vector<float> hist) {
	float ret = 0.0;
	for(unsigned int i = 0; i < hist.size(); i++) {
		ret += i * hist[i];
	}
	return ret;
}


/**************************************************************************************************

	Returns the variance

**************************************************************************************************/
float GetHistogramVariance(std::vector<float> hist, float avg) {
	float ret = 0.0;
	for(unsigned int i = 0; i < hist.size(); i++) {
		ret += pow((i - avg), 2) * hist[i];
	}

	return ret;
}

/**************************************************************************************************

	Returns the skewness

**************************************************************************************************/
float GetHistogramSkewness(std::vector<float> hist, float avg, float variance) {
	float ret =0.0;
	for(unsigned int i = 0; i < hist.size(); i++) {
		ret += pow((i - avg), 3) * hist[i];
	}
	// var^2  -> var^3
	variance = variance * sqrt(variance);
	ret /= variance;
	return ret;
}


/**************************************************************************************************

	Returns the kurtosis

**************************************************************************************************/
float GetHistogramKurtosis(std::vector<float> hist, float avg, float variance) {
	float ret = 0.0;
	for(unsigned int i = 0; i < hist.size(); i++) {
		ret += pow((i - avg), 4) * hist[i];
	}
	// var^2  -> var^4
	variance = pow(variance,2);
	ret /= variance;
	ret -= 3;
	return ret;
}


/**************************************************************************************************

	Computes the variation coefficient

**************************************************************************************************/
float GetVarianceCoeff(float variance, float average) {
	float stddev = sqrt(variance);
	return stddev/average;
}


/**************************************************************************************************

	Log_2 function

**************************************************************************************************/
float Log2(float x) {
	// things fuck up when X is 0. check with someone that this is correct.
	if(x == 0)
		return 0;
	float ret = log10f(x) / log10f(2);
	return ret;
}


/**************************************************************************************************

	Computes the Entropy for an image from its histogram

**************************************************************************************************/
float GetEntropy(std::vector<float> histogram) {
	float entropy = 0.0;
	for(unsigned int i  = 0; i < histogram.size(); i++) {
		entropy += histogram[i] * Log2(histogram[i]);
	}
	entropy *= -1;

	return entropy;
}


/************************************************************************************************

	Compute centroid of segment(s)

************************************************************************************************/
template<class ImageType>
typename ImageType::IndexType GetCentroids(typename ImageType::Pointer image) {
	
	float xa = 0;
	float ya = 0;
	float za = 0;
	float counter = 0;
	ImageType::SizeType size = GetSizeOfImage<ImageType>(image);
	for(int x = 0; x<size[0]; x++) {
		for(int y = 0; y<size[1] ; y++) {
			for(int z =0; z<size[2] ; z++) {
				ImageType::IndexType index; index[0] = x, index[1] = y, index[2] = z;	
				if(image->GetPixel(index) > 0) {
					xa += x;
					ya += y;
					za += z;
					counter++;
				}
			}
		}
	}
//	std::cout << xa << " " << ya << " " << za << " " << counter << std::endl;
	xa /= counter;
	ya /= counter;
	za /= counter;

	ImageType::IndexType index;
	index[0] = (int)xa;
	index[1] = (int)ya;
	index[2] = (int)za;

//	image->SetPixel(index, 127);

	return index;

}



/*************************************************************************************************************************************

	Returns centroids with respect to color. 

*************************************************************************************************************************************/
template<class ImageType>
std::vector<float> GetCentroidsColor(typename ImageType::Pointer image, float average) {
	
	float xa = 0;
	float ya = 0;
	float za = 0;
	float counter = 0;
	ImageType::SizeType size = GetSizeOfImage<ImageType>(image);
	float S = size[0] * size[1] * size[2];

	for(int x = 0; x<size[0]; x++) {
		for(int y = 0; y<size[1] ; y++) {
			for(int z =0; z<size[2] ; z++) {
				ImageType::IndexType index; index[0] = x, index[1] = y, index[2] = z;	
//				std::cout << index << " har " << image->GetPixel(index) << std::endl;
				xa += x * (image->GetPixel(index) / (average*S));
				ya += y * (image->GetPixel(index) / (average*S));
				za += z * (image->GetPixel(index) / (average*S));
			}
		}
	}

//	std::cout << xa << " " << ya << " " << za << " " << S << " Average " << average << std::endl;

	std::vector<float> index;
	index.push_back(xa);
	index.push_back(ya);
	index.push_back(za);

	return index;

}

template<class ImageType>
int GetTotal(typename ImageType::Pointer img) {
	int sum = 0;
	typedef itk::ImageRegionConstIterator<ImageType> itType;
	itType it(img, img->GetLargestPossibleRegion());
	while(!it.IsAtEnd()) {
		sum += it.Get();
		++it;
	}
	return sum;
}

/************************************************************************************************************************

		Part of the moments-things. 

************************************************************************************************************************/
template<class ImageType>
float mpqr(typename ImageType::Pointer image, int p, int q, int r) {
	ImageType::SizeType size = GetSizeOfImage<ImageType>(image);
	float m = 0;
	float total = GetTotal<ImageType>(image);

	ImageType::IndexType centroid = GetCentroids<ImageType>(image);

	for(int x = 0; x<size[0]; x++) {
		for(int y = 0; y<size[1] ; y++) {
			for(int z =0; z<size[2] ; z++) {
				ImageType::IndexType index; index[0] = x, index[1] = y, index[2] = z;	
//				if(image->GetPixel(index) == 255)
//					m += pow(x,p) * pow(y,q) * pow(z,r);
					m += pow((x-centroid[0]),p) * pow((y-centroid[1]),q) * pow((z-centroid[2]),r) * (((float)(image->GetPixel(index)))/total);
			}
		}
	}
//	std::cout << "m"<<p<<","<<q<<","<<r<<" = " << m << endl;
	return m;
}



/*********************************************************************************************************************

	Moments.

*********************************************************************************************************************/
template<class ImageType>
float OrderOne(typename ImageType::Pointer image) {
	float ret = 0;
	ret = mpqr<ImageType>(image, 2,0,0) + mpqr<ImageType>(image,0,2,0) + mpqr<ImageType>(image,0,0,2);
	return ret;
}

template<class ImageType>
float OrderTwo(typename ImageType::Pointer image) {
	float ret = 0;
	ret = mpqr<ImageType>(image, 2,0,0) * mpqr<ImageType>(image,0,2,0) + 
		mpqr<ImageType>(image,2,0,0) * mpqr<ImageType>(image,0,0,2) +
		mpqr<ImageType>(image,0,2,0) * mpqr<ImageType>(image,0,0,2) -
		pow(mpqr<ImageType>(image,1,1,0),2) -
		pow(mpqr<ImageType>(image,1,0,1),2) -
		pow(mpqr<ImageType>(image,0,1,1),2); 
	return ret;
}

template<class ImageType>
float OrderThree(typename ImageType::Pointer image) {
	float ret = 0;
	ret = mpqr<ImageType>(image,2,0,0) * mpqr<ImageType>(image,0,0,2) * mpqr<ImageType>(image,0,0,2) +
		2*mpqr<ImageType>(image,1,1,0) * mpqr<ImageType>(image,1,0,1) * mpqr<ImageType>(image,1,1,0) -
		mpqr<ImageType>(image,0,0,2)*pow(mpqr<ImageType>(image,1,1,0),2) -
		mpqr<ImageType>(image,0,2,0)*pow(mpqr<ImageType>(image,1,0,1),2) -
		mpqr<ImageType>(image,2,0,0)*pow(mpqr<ImageType>(image,0,1,1),2);

	return ret;
}


/************************************************************************************************************************

	Computes and returns the three moments

************************************************************************************************************************/
template<class ImageType>
std::vector<float> GetMoments(typename ImageType::Pointer image) {
//	GetCentroidsColor<ImageType>(image, 4);
	std::vector<float> retVector;
	retVector.push_back(OrderOne<ImageType>(image));
	retVector.push_back(OrderTwo<ImageType>(image));
	retVector.push_back(OrderThree<ImageType>(image));

	

	return retVector;
}




/*

		Will extract moments and statistics on the region currently believed to be ventricles

*/
template<class ImageType>
std::vector<float> GetInPlaceStatistics(typename ImageType::Pointer segmented, typename ImageType::Pointer original) {
	ImageType::Pointer useImg = CBRCopyImage<ImageType>(original);

	typedef itk::ImageRegionIterator<ImageType> itType;
	itType itSeg(segmented, segmented->GetLargestPossibleRegion());
	itType itUse(useImg, useImg->GetLargestPossibleRegion());
	itSeg.GoToBegin();
	itUse.GoToBegin();

	while(!itSeg.IsAtEnd()) {
		if(itSeg.Get() != 255) {
			itUse.Set(0);
		}
		++itSeg;
		++itUse;
	}

	WriteImageFile<ImageType>(useImg, "test.mhd");
	std::vector<float> retVector = GetHistogramStatistics<ImageType>(useImg);
	std::vector<float> tmp = GetMoments<ImageType>(useImg);
	for(unsigned int i = 0; i< tmp.size(); i++)
		retVector.push_back(tmp[i]);

	//for(unsigned int i = 0; i< retVector.size(); i++)
	//	std::cout << retVector[i] << std::endl;

	return retVector;
}


#endif