#include "CBRWindow.h"
#include "xml2cbr.h"
#include "PLogic.h"

#include <QWidget>
#include <QPushButton>
#include <QApplication>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QComboBox>
#include <QString>
#include <QLabel>
#include <vector>
#include <QLineEdit>
#include <QIntValidator>
/*

	Sets up the window for registration of new case.
*/

 QWidget* createInputWindow(xml2cbr &cbrlogic, PLogic &mainWindow) {
	QWidget *window = new QWidget(mainWindow);

	QVBoxLayout *layout = new QVBoxLayout();
	QHBoxLayout *hlayout = new QHBoxLayout();

	QPushButton * quit = new QPushButton("Quit", window);
	connect(quit, SIGNAL(clicked()), qApp, SLOT(quit()));
	
	QPushButton * match = new QPushButton("Match", window);
	connect(match, SIGNAL(clicked()), mainWindow, SLOT(match()));
	
	// iterator to go through
	map<std::string, CBRFeature> feats = cbrlogic.getFeatures();
	map<std::string, CBRFeature>::const_iterator iter;

	// add one ComboBox for each allowable index
	for(iter = feats.begin(); iter != feats.end(); ++iter) {
		CBRFeature f = iter->second;

		// for each label/input pair, place in a horizontal box.
		QHBoxLayout *ql = new QHBoxLayout();

		QLabel *label = new QLabel(QString(f.getName().c_str()), window);
		ql->addWidget(label);

		
		

		// if the input-field is numeric, get text and validate.
		if(f.isOfTypeInt()) {
			QIntValidator *ivalid = new QIntValidator(f.getMin(), f.getMax(), window);
			QLineEdit *inputline = new QLineEdit(window);
			inputline->setValidator(ivalid);
			inputline->setFixedWidth(200);
			inputline->setName(f.getName().c_str());
			ql->addWidget(inputline);

			connect(inputline, SIGNAL(editingFinished()), mainWindow, SLOT(inputIntChanged()));
		}
		// will be float or isOfTypeInt would kicked in
		else if(f.isNumeric()) {
			//create input field and a validator. 
			QDoubleValidator *dvalid = new QDoubleValidator(f.getMin(), f.getMax(),2, window);
			QLineEdit *inputline = new QLineEdit(window);
			inputline->setValidator(dvalid);
			inputline->setFixedWidth(200);
			inputline->setName(f.getName().c_str());
			ql->addWidget(inputline);

			connect(inputline, SIGNAL(editingFinished()), mainWindow, SLOT(inputIntChanged()));
		}
		// string, make a drop down box
		else {

			QComboBox *value = new QComboBox(window);
			value->setName(QString(f.getName().c_str()));
			value->setFixedWidth(200);
			//get name from CBRFeature
			std::vector<std::string> fvalues = f.getAllowedValues();

			// iterate through f to get allowed values
			for(unsigned int fsize = 0; fsize < fvalues.size(); fsize++ ) {
				std::string fstringtmp = fvalues.at(fsize);
				QVariant *var = new QVariant(QString(fstringtmp.c_str()));
				value->addItem(QString(fstringtmp.c_str()), *var);
			}

			connect(value, SIGNAL(currentIndexChanged(int)), mainWindow, SLOT(ddboxChanged(int)));
			// save to window
			ql->addWidget(value);
		}
		layout->addLayout(ql);
	}

	//position things
	hlayout->addWidget(quit);
	hlayout->addSpacing(30);
	hlayout->addWidget(match);

	layout->addLayout(hlayout);
	setLayout(layout);
	return window;
 }
