#include <itkImageFileReader.h>
#include <itkImageFileWriter.h>
#include <itkCastImageFilter.h>
#include <itkImage.h>
#include <itkCurvatureFlowImageFilter.h> 
#include <itkDataObject.h>
#include <itkNeighborhoodConnectedImageFilter.h>
#include <itkRescaleIntensityImageFilter.h>
#include <itkIsolatedConnectedImageFilter.h>
#include <itkImageRegion.h>

#include <climits>
#include <utility>
#include <fstream>
//#include <itkSpatialOrientation.h>
#include <itkOrientImageFilter.h>

#include <itkVersorRigid3DTransform.h>
#include <itkCenteredTransformInitializer.h>
#include <itkVersorRigid3DTransformOptimizer.h>

#include <itkLinearInterpolateImageFunction.h>
#include <itkMattesMutualInformationImageToImageMetric.h>
//#include <itkVersorTransformOptimizer.h>

#include "CBRitk.h"
#include "CBRImageToImageComparators.h"
#include "CBRStatistics.h"
#include "Registrate.h"
#include "CBRImageFunctions.h"
#include "CBRLevelSet.h"
#include "CBRCannyLevelset.h"
#include "CBRCase.h"


//const float VentricleValue = 128.5078;
const int VentricleValue = 128;
const int Dimension = 3;

	// keep the best during registration
itk::Image<float, 3>::Pointer bestImg = 0;
itk::Image<float, 3>::Pointer bestSol = 0;

/*
	
	Set some flags for runtime tweaking

*/
void setFlags(int argc, char *argv[]) {
	for(int i = 0; i < argc; i++) {
		std::string tmp = argv[i];
		if(tmp.substr(0,1) == "-") {
			std::cout << tmp << std::endl;
			char key = tmp.substr(1,1).c_str()[0];
			switch(key) {
			case 'q':
				quickie=true;
				break;
			}
		}
	}
}


/*
	Takes the full name (+path) of an image file and loads it.
	It reorients the file and rescales it to 255 before
	returning a T::Pointer.
*/
template<class T>
typename T::Pointer GetImageFile(std::string file) {
	//typedefs

	typedef itk::Image<float, Dimension> inputImageType;	
	typedef itk::ImageFileReader<inputImageType> ReaderType;
	typedef itk::CastImageFilter<inputImageType, T> CasterType;
	typedef itk::OrientImageFilter<T, T> OrientType;
	typedef itk::RescaleIntensityImageFilter<T,T> filterType;

	// instantiate
	CasterType::Pointer caster = CasterType::New();
	ReaderType::Pointer reader = ReaderType::New();
	OrientType::Pointer orient = OrientType::New();
	filterType::Pointer rescale = filterType::New();

	// pipeline
	reader->SetFileName(file);
	caster->SetInput(reader->GetOutput());
	orient->SetInput(caster->GetOutput());
	rescale->SetInput(orient->GetOutput());

	orient->UseImageDirectionOn();
	orient->SetDesiredCoordinateOrientationToAxial();

	rescale->SetOutputMaximum(255);
	rescale->SetOutputMinimum(0);

	T::Pointer ret = T::New();
	ret = rescale->GetOutput();
	//ret = caster->GetOutput();
	//T::Pointer image = rescale->GetOutput();


//	WriteImageFile<ImageType>(orient->GetOutput(), "toAxial.mhd");
	try{
		ret->Update();
	}
	catch(itk::ExceptionObject &exp) {
		std::cerr << exp << std::endl; 
		std::cerr << "feil ved lasting av " << file <<std::endl;
	}

	std::cout << "read " << file << std::endl;
	return ret;
}



/*
	Takes a T::Pointer to a file and the name of the file to be written.
	takes the cbrcase and forwards to writeVVI, who needs parameters.
	*/
template<class T>
void WriteImageFile(typename T::Pointer image, std::string imgName, CBRCase *outCase = 0) {
	typedef itk::Image<unsigned char, Dimension> outputType;
	typedef itk::ImageFileWriter<outputType> WriterType;

	WriterType::Pointer writer = WriterType::New();

	typedef itk::CastImageFilter<T, outputType> CasterType;
	typedef itk::RescaleIntensityImageFilter<outputType,outputType> filterType;

	CasterType::Pointer caster = CasterType::New();
	filterType::Pointer rescale = filterType::New();

	caster->SetInput(image);
	rescale->SetInput(caster->GetOutput());
//	writer->SetInput(rescale->GetOutput());	
	writer->SetInput(caster->GetOutput());
	rescale->SetOutputMaximum(255);
	rescale->SetOutputMinimum(0);

	writer->SetFileName(imgName);
	T::SizeType size = GetSizeOfImage<T>(image);

	outputType::Pointer toFile = rescale->GetOutput();
	try{writer->Update();}
	catch(itk::ExceptionObject &exp) {
		std::cerr << exp;
		std::cerr << "prvde  skrive: " << imgName << std::endl;
		exit(1);
	}
	std::cout << "Wrote " << imgName << " with size " << size << std::endl;

	// write VVI?
//	if(outCase != NULL)
	toFile->Update();
	outputType::SizeType vviSize = GetSizeOfImage<outputType>(toFile);
	WriteVVI(imgName, vviSize[0], vviSize[1], vviSize[2]);
}



//COPIED FROM INTERNET --- START 
// http://www.codeguru.com/forum/showthread.php?s=&threadid=231054
template <class T>
bool from_string(T& t, 
				 const std::string& s, 
				 std::ios_base& (*f)(std::ios_base&))
{
	std::istringstream iss(s);
	return !(iss >> f >> t).fail();
}

// COPIED FOR INTERNET --- END

template<class T>
void PrintStats(typename T::Pointer image)
{
	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;
	image->Update();
	ConstIteratorType it(image, image->GetRequestedRegion());

	float fMin = std::numeric_limits<float>::max(); 
	float fMax = std::numeric_limits<float>::min();
	float fAvg = 0;
	size_t nCtr = 0;

	it.GoToBegin();
	while (!it.IsAtEnd())
	{
		float fVal = it.Get();
		if (fVal != 0)
		{
			fAvg += fVal;
			nCtr++;
		}
		if (fVal < fMin)
			fMin = fVal;
		if (fVal > fMax)
			fMax = fVal;
		++it;
	}

	std::cout << "Min: " << fMin << ", max: " << fMax << ", Avg: " << fAvg / nCtr << " (ctr: " << nCtr << ").\n";
}
/*




/*
	Creates an overlap between image and the segmented file fileName.
*/
template<class T>
typename T::Pointer createMatchImage(typename T::Pointer image, std::string fileName) {

	std::cout << "Matching solutions " <<std::endl;
	//iterators
	typedef itk::ImageRegionConstIterator<T> ConstIteratorType;
	typedef itk::ImageRegionIterator<T > IteratorType;


	// read solution image and output image
	T::Pointer solImage = GetImageFile<T>(fileName);
	T::Pointer mImage = image;
	solImage->Update();
	mImage->Update();


	//create iterators.
	ConstIteratorType it(image, image->GetRequestedRegion());
	ConstIteratorType itS(solImage, solImage->GetRequestedRegion());
	IteratorType itM(mImage, mImage->GetRequestedRegion());

	// put together overlap-image.
	int corr = 0;
	int fp = 0;
	int fn = 0;
	itS.GoToBegin();
	itM.GoToBegin();
	it.GoToBegin();
	while(!it.IsAtEnd()) {
		float fVal = it.Get();
		float sVal = itS.Get();
		if(fVal == 255) {
			if(sVal == VentricleValue) {
				//direct match
				itM.Set(255);
				corr++;
			}
			else {
				//false positive
				itM.Set(100);
				fp++;
			}
		}
		// false negative
		else {
			if(sVal == VentricleValue) {
				itM.Set(190);
				fn++;
			}
			else {
				// no match
				itM.Set(0);
			}
		}
		++itS;
		++itM;
		++it;
	}

//	std::cout << "Correct: " << corr << " Missed: "<< fn << " Wrong: " << fp << std::endl;
//	float percentage = (float)(corr+fn) / (float)(corr+fp);
//	std::cout << percentage << std::endl;

	return mImage;
}





template<class T>
void printSize(typename T::Pointer image) {

	typedef itk::ImageRegion<Dimension> RegionType;

	RegionType region = image->GetLargestPossibleRegion();	
	RegionType::SizeType size = region.GetSize();

	std::cout << size << std::endl;

	//std::cout << "x " << size[0] << std::endl;
	//std::cout << "y " << size[1] << std::endl;
	//std::cout << "z " << size[2] << std::endl;
}

template<class T>
typename T::Pointer RescaleImage(typename T::Pointer image, int min, int max) {
	
	typedef itk::RescaleIntensityImageFilter<T,T> filterType;
	filterType::Pointer rescale = filterType::New();

	rescale->SetOutputMaximum(max);
	rescale->SetOutputMinimum(min);
	rescale->SetInput(image);

	T::Pointer ret = T::New();
	ret = rescale->GetOutput();

	try{ 	rescale->Update(); ret->Update(); }
	catch(itk::ExceptionObject & exp) { std::cerr << exp << std::endl; exit(0); }
	return ret;

}




/**************************************************************************************

	The function that does the final thresholding 
	Uses the thresholds and coordinates found in incomingcase.
	Adaption is expected done prior to this step.

**************************************************************************************/
template<class ImageType>
typename ImageType::Pointer UseThreshold(CBRCase &incomingCase, std::string original, int round=0) {
//	std::string saveDir = "out\\Brute\\" + incomingCase.GetStringIndex("file") + "\\";
	std::string saveDir = "out\\revise\\";
	std::string utDir = "";

	typedef itk::Image<int, Dimension> ImageType;
	

	ImageType::Pointer orig;
	ImageType::Pointer seg;

	if(original == "old") {
	//build string with image names and get images
		std::string segString = "seg\\" + incomingCase.GetStringIndex("file") +".hdr"; 
		std::string origString = "orig\\" + incomingCase.GetStringIndex("file") + ".hdr";
		orig = GetImageFile<ImageType>(origString);
		seg = GetImageFile<ImageType>(segString);
	}
	else {
		std::cout << "casting " << std::endl;
		orig = CastImage<float, int>(bestImg);
		seg = CastImage<float, int>(bestSol);
		utDir = saveDir +"arriving.mhd";
//		WriteImageFile<ImageType>(orig, utDir, &incomingCase);
	}
	//get subvolue
	ImageType::IndexType min;
	ImageType::IndexType max;
	int revise = 0;

	min[0] = incomingCase.GetIntIndex("Xmin");
	min[1] = incomingCase.GetIntIndex("Ymin");
	min[2] = incomingCase.GetIntIndex("Zmin");
	max[0] = incomingCase.GetIntIndex("Xmax");
	max[1] = incomingCase.GetIntIndex("Ymax");
	max[2] = incomingCase.GetIntIndex("Zmax");

	// get subvolume. Save to disk for inspection.
	ImageType::Pointer subimage = getSubvolume<ImageType>(orig, min, max);
	ImageType::Pointer subsol = getSubvolume<ImageType>(seg, min, max);

	std::cout << "-----------------------------NEW ROUND---------------------------"<<std::endl;

	//ImageType::Pointer tmp =  Threshold<ImageType>(subimage, incomingCase.GetIntIndex("firstThresh"));
	//tmp = keepBiggestRegions<ImageType>(tmp, 1);
	//ImageType::IndexType cent = GetCentroids<ImageType>(tmp);
	//std::cout << "First centroid: " << cent << std::endl;
	//tmp->SetPixel(cent, 128);
	//WriteImageFile<ImageType>(tmp, "out\\revise2\\first.mhd");
	//AdaptDebug<ImageType>(subimage, orig);

	//remove the biggest region remaining after thresholding the image at 
	// argv[4]. actually set it to 255.
	ImageType::Pointer blacked = blackOut<ImageType>(subimage, incomingCase.GetIntIndex("firstThresh"));

	// write out result after largest region has been removed - debug
	utDir = saveDir + ToString(round)+ "-white-.mhd";
	WriteImageFile<ImageType>(blacked, utDir, &incomingCase);

	// threshold second time
	ImageType::Pointer done = Threshold<ImageType>(blacked, incomingCase.GetIntIndex("secondThresh"));

	// only keep the largest x regions of threshold
	done = keepBiggestRegions<ImageType>(done, 1);

	// write out result of thresholding
	utDir = saveDir +  ToString(round)+ "-thresh-debug.mhd";
	WriteImageFile<ImageType>(done, utDir, &incomingCase);

	// get moments - used in debug
	std::vector<float> momentsVector = GetMoments<ImageType>(done);
	std::cout << "Momenter, lite bilde " << std::endl;
	//std::cout << momentsVector[0] << "," << momentsVector[1] << "," << momentsVector[2] << std::endl;
	std::cout << "\t\t\t\t<Moment1>" << momentsVector[0] << "</Moment1>"<< std::endl;
	std::cout << "\t\t\t\t<Moment2>" << momentsVector[1] << "</Moment2>"<< std::endl;
	std::cout << "\t\t\t\t<Moment3>" << momentsVector[2] << "</Moment3>"<< std::endl;


	ImageType::IndexType cen = GetCentroids<ImageType>(done);
	std::cout << cen << std::endl;

	// merge thresholded subimage with large image again
	MergeImages<ImageType>(orig, done, min);
	utDir = saveDir + "subsol.mhd";
//	WriteImageFile<ImageType>(subsol, utDir);

	std::cout << "Full score: ";
	getScore<ImageType>(orig, seg);

	// get score of threshold.
	getScore<ImageType>(done, subsol, incomingCase.GetIntIndex("firstThresh"), incomingCase.GetIntIndex("secondThresh"));
//	Revise<ImageType>(done, incomingCase,0);
	GetInPlaceStatistics<ImageType>(done, subimage);

	// create overlap image and write it to disk.
	ImageType::Pointer smallMatch = createMatchImageToImage<ImageType>(done, subsol);

	utDir = saveDir +  ToString(round)+ "-overlap-debug.mhd";
	WriteImageFile<ImageType>(smallMatch, utDir, &incomingCase);

	//// create overlap image and write it to disk.
	seg = createMatchImageToImage<ImageType>(orig, seg);
	utDir = saveDir + "big-overlap-debug.mhd";
	WriteImageFile<ImageType>(seg, utDir);


	return done;
}










void SplitAndThreshold(int argc, char *argv[]) {
	typedef itk::Image<float, Dimension> ImageType;

	// threshold to use. Easier to remember than argv[something]
	int threshold = atoi(argv[3]);

	// read files from disk. FS = fullsize
	ImageType::Pointer originalFS = GetImageFile<ImageType>(argv[1]);
	ImageType::Pointer solutionFS = GetImageFile<ImageType>(argv[2]);
	ImageType::Pointer toSpeed = GetImageFile<ImageType>("D:\\Diplom\\code\\seg\\17_3.hdr");
	ImageType::Pointer fixed = GetImageFile<ImageType>("D:\\Diplom\\code\\orig\\17_3.hdr");
//	printVentricleBoundary<ImageType>(solutionFS);

	//return from registrate
	std::pair<ImageType::Pointer, ImageType::Pointer> transformedPair;
	//registrate the volumes with ventricle, but transform the full image.
	//transformedPair	= Registrate<ImageType>(fixed, originalFS, solutionFS, originalFS, solutionFS);
	//originalFS = transformedPair.first;
	//solutionFS = transformedPair.second;

	//transformedPair	= Registrate<ImageType>(fixed, original, solution, originalFS, solutionFS);
	//originalFS = transformedPair.first;
	//solutionFS = transformedPair.second;

	// debug
	//WriteImageFile<ImageType>(originalFS, "registreredO.mhd");
	//WriteImageFile<ImageType>(solutionFS, "registreredS.mhd");

	ImageType::Pointer original = originalFS;
	ImageType::Pointer solution = solutionFS;
	// calculate speed image from the solution-file of case-image. 


	// extract only ventricle and use as initial model
	ImageType::Pointer speed = createFastMarchingImage<ImageType>(toSpeed);
//	ImageType::Pointer speed = toSpeed;
	
	//	WriteImageFile<ImageType>(speed, "speed.mhd");
//	WriteImageFile<ImageType>(solution, "solution.mhd");
//	PrintStats<ImageType>(speed);

	
//	ImageType::Pointer ret = CannyLevelset<ImageType>(original, speed, atof(argv[3]), atoi(argv[4]), atoi(argv[5]));
	std::cout << "Beginning Levelset method" << std::endl << std::flush;
	ImageType::Pointer ret = LaplacianCanny<ImageType>(original, speed, atoi(argv[3]), atof(argv[4]), atof(argv[5]), atof(argv[6]), atof(argv[7]));

	std::cout << "Returned " << std::endl;
	WriteImageFile<ImageType>(ret, "out\\debug\\levelset.mhd");
	WriteImageFile<ImageType>(original, "out\\debug\\orig.mhd");


	//temporary
	WriteImageFile<ImageType>(speed, "out\\debug\\speed.mhd");
	exit(1);
	

	// get score of threshold.
	getScore<ImageType>(ret, solution);

	// create overlap image and write it to disk.
	ImageType::Pointer subsol = createMatchImageToImage<ImageType>(ret, solution);
	std::string overlapBase = "D:\\Diplom\\Code\\out\\overlap-";
	//overlapBase += argv[3];
	overlapBase += "canny";
	overlapBase += ".mhd";
	WriteImageFile<ImageType>(subsol, overlapBase);

}


void Checkbounds(int argc, char *argv[]) {
	typedef itk::Image<int, Dimension> ImageType;

	// threshold to use. Easier to remember than argv[something]
	int threshold = 0;

	// read files from disk. FS = fullsize
	ImageType::Pointer originalFS = GetImageFile<ImageType>(argv[1]);
	ImageType::Pointer solutionFS = GetImageFile<ImageType>(argv[2]);


	// used to house which regions are to survive
	std::vector<int> allowed;
	allowed.push_back(2);

	// hard coded for 1_24
	ImageType::IndexType min;
	ImageType::IndexType max;
	min[0] = atoi(argv[4]); // 100
	min[1] = atoi(argv[5]); // 23;
	min[2] = atoi(argv[6]); //117;
	max[0] = atoi(argv[7]); //163;
	max[1] = atoi(argv[8]); //60;
	max[2] = atoi(argv[9]); //145;

	// get subvolume. Save to disk for inspection.
	ImageType::Pointer original = getSubvolume<ImageType>(originalFS, min, max);
	ImageType::Pointer solution = getSubvolume<ImageType>(solutionFS, min, max);

	std::string sunBase = "D:\\Diplom\\Code\\ut\\sub";
	sunBase += ".mhd";
	WriteImageFile<ImageType>(solution, sunBase);
	printSize<ImageType>(solution);
	
	exit(1);

	// threshold the small subvolume
	ImageType::Pointer threshed = Threshold<ImageType>(originalFS, VentricleValue);
		
	threshed = keepBiggestRegions<ImageType>(threshed, 2);

	// create overlap image and write it to disk.
//	ImageType::Pointer subsol = createMatchImageToImage<ImageType>(threshed, solution);
	std::string overlapBase = "D:\\Diplom\\Code\\out\\overlap-";
//	overlapBase += argv[3];
	overlapBase += "debug";
	overlapBase += ".mhd";
	WriteImageFile<ImageType>(threshed, overlapBase);
	printSize<ImageType>(threshed);
}





/**************************************************************************************

	Writes the .VVI used by VolView to allow easy loading of the images.
	Save a lot of time in inspection of result, does not affect how the 
	program actually works.

**************************************************************************************/

// write the .vvi file used by VolView.
void WriteVVI(std::string filename, int x, int y, int z) {
	// name.mhd -> name.raw.vvi
	filename = Split(filename, '.')[0];
	filename += ".raw.vvi";
	ofstream file(filename.c_str());

	//write
	file << "<VVOpenWizard Version=\"1.8\" \n ";
	file << "Spacing=\"1 3.1 1\" \n";
	file << "Origin=\"0 0 0\" \n";
	file << "ScalarType=\"3\" \n";


	//write
	file << "WholeExtent=\"0 "<< (x-1) << " 0 "<< (y-1) <<" 0 " << (z-1) <<"\" \n";
	file << "NumberOfScalarComponents=\"1\" \n";
	file << "IndependentComponents=\"1\" \n";
	file << "FileOrientation=\"10 2 0\" \n";
	file << "BigEndianFlag=\"0\" \n";

	// string manipulation.
	// .hdr -> .raw
	std::string patternName = (Split(filename, '\\')).back();
	patternName = (Split(patternName, '.'))[0];
	patternName += ".raw";

	//write
	file << "FilePattern=\""<< patternName <<"\" \n";
	file << "FileDimensionality=\"3\" />";

	//close file
	file.close();
}



/************************************************************************************************/


// returns a vector of strings, the input string split at the char splitter. ( see phps explode)
std::vector<std::string> Split(std::string toSplit, char splitter) {
	std::string tmpString;
	std::vector<std::string> retVector;
	bool copy = false;
	for(unsigned int i = 0; i< toSplit.size() ; i++) {
		if(toSplit[i] == splitter) {
			retVector.push_back(tmpString);
			tmpString = "";
		}
		else {
			tmpString += toSplit[i];
		}
	}
	if(tmpString != "")
		retVector.push_back(tmpString);
	
	return retVector;
}





/*********************************************************************************

	Matching function.
	Accepts a vector of previous cases and a newCase.
	Will load images from disk, registrate them and get the score.
	The registration with the lowest score will be saved in 
	the global itk::Image<float,3>::Pointers bestImg and bestSol.


*********************************************************************************/
CBRCase MatchCases(std::vector<CBRCase> cases, CBRCase &newcase) {

	// type of image to work on, and the images from newcase
	typedef itk::Image<float, Dimension> ImageType;
	std::string newfile = "orig\\" + newcase.GetStringIndex("file") + ".hdr";
	ImageType::Pointer newImg = GetImageFile<ImageType>(newfile);
	std::string solstr = "seg\\" + newcase.GetStringIndex("file") + ".hdr";
	ImageType::Pointer newsol = GetImageFile<ImageType>(solstr);


	std::string forbidden = newcase.GetStringIndex("file");

	// score is bound by reference in registrate. 
	float score = 0.0;
	float best = std::numeric_limits<float>::max(); 
	CBRCase *bestCase = 0;

	// -q flag from command prompt. just juse the best case.
	if(quickie) {
		std::string quickUse;
		for(unsigned int i = 0; i< cases.size() ; i++) {
			if(cases[i].GetStringIndex("file") == forbidden) {
				std::pair<typename ImageType::Pointer, typename ImageType::Pointer> scorePair;
				scorePair = GetImageDistance<ImageType>(cases[i].GetStringIndex("use"), newImg, newsol, score);
				bestImg = scorePair.first;
				bestSol = scorePair.second;
				quickUse = cases[i].GetStringIndex("use");
				break;
			}
		}
		for(unsigned int i = 0; i< cases.size() ; i++) {
			if(cases[i].GetStringIndex("file") == quickUse) {
				bestCase = &cases[i];
				break;
			}
		}
	}
	else {
		for(unsigned int i = 0; i< cases.size() ; i++) {
			// do not registrate to your self
			if(cases[i].GetStringIndex("file") == forbidden || cases[i].GetStringIndex("skip") == "yes") {
				continue;
			}
			else {
				std::cout << "Registrerer" << std::endl;
				std::pair<typename ImageType::Pointer, typename ImageType::Pointer> scorePair;
				scorePair = GetImageDistance<ImageType>(cases[i].GetStringIndex("file"), newImg, newsol, score);
				if(score < best) {
					std::cout << "New best, from "<< best << " to " << score << " case "<< cases[i].GetStringIndex("file") << std::endl;
					bestImg = scorePair.first;
					bestSol = scorePair.second;
					best = score;
					bestCase = &cases[i];		
				}
			} // else
		}
	}

	std::cout << "Beste distanse er: " << best << " med " << bestCase->GetStringIndex("file") << std::endl;
	return *bestCase;
}



/*********************************************************************************************************

	Takes two ImagePointers, the original and solution of the new case.
	This is registrered to the old case and the registrered versions along with the 
	final distance is returned.

*********************************************************************************************************/
template<class ImageType>
std::pair<typename ImageType::Pointer, typename ImageType::Pointer> GetImageDistance(std::string fixed, 
																					 typename ImageType::Pointer moving,
																					 typename ImageType::Pointer movSol,
																					 float &score) {
			 ImageType::Pointer moving2 = CBRCopyImage<ImageType>(moving);
			 ImageType::Pointer movSol2 = CBRCopyImage<ImageType>(movSol);
			// load the moving image from new case, from disk.
			std::string fixedfile = "orig\\" + fixed + ".hdr";
			ImageType::Pointer fixedImg = GetImageFile<ImageType>(fixedfile);

			// the calculation, keep the registrered image in the pair
			std::pair<typename ImageType::Pointer, typename ImageType::Pointer> regPair;

//			WriteImageFile<ImageType>(movSol2, "out\\solPre.mhd");
			regPair = Registrate<ImageType>(fixedImg, moving2, movSol2, score);
//			WriteImageFile<ImageType>(regPair.second, "out\\solPost.mhd");

			return regPair;
}





/************************************************************************************************

	Entry point for processing the chosen CBRCase and the newCase

************************************************************************************************/
void ProcessCase(CBRCase &oldCase, CBRCase &newCase) {
	
	//adapt the incoming solution.
	Adapt(oldCase, newCase);

	// process the adapted new case.
	Loop3(newCase, oldCase, "new");	
}



void balle(std::string a) {
	typedef itk::Image<int, 3> ImageType;
	std::string bilde = "orig\\" + a;
	ImageType::Pointer img = GetImageFile<ImageType>(a);

	ImageType::Pointer ut = Threshold<ImageType>(img, 116);

	WriteImageFile<ImageType>(ut, "out\\debug\\singleThr.mhd");
}





/************************************************************************************************************





	Similarity function for statistical moments


*************************************************************************************************************/
CBRCase MatchStatistics(std::vector<CBRCase> cases, CBRCase &newcase) {
	float tmp = 0.0;
	float score = std::numeric_limits<float>::max();
	std::vector<float> newCaseStats = GetStatisticsFromNewImage(newcase.GetStringIndex("file"));

	CBRCase *bestCase = 0;
	std::vector<std::string> indices;
	indices.push_back("AverageFull"), indices.push_back("VarianceFull"), indices.push_back("SkewnessFull"), indices.push_back("KurtosisFull");
	indices.push_back("CoeffVariationFull"),indices.push_back("EntropyFull");
	indices.push_back("CentroidXFull"), indices.push_back("CentroidYFull"),	indices.push_back("CentroidZFull");

	//iterate cases
	for(int j = 0; j < cases.size() ; j++) {
		if(cases[j].GetStringIndex("file") == newcase.GetStringIndex("file"))
			continue;

		//iterate for all indices
		for(int i = 0; i < indices.size() ; i++) {
			std::pair<float, float> minmax = GetMinMaxForAttribute(indices[i], cases);
			tmp += GetStatisticalDistance(newCaseStats[i], cases[j].GetFloatIndex(indices[i]), minmax.first, minmax.second);
		}
		tmp /= ((float)indices.size());

		if(tmp < score)  {
			score = tmp;
			bestCase = &cases[i];		
//			std::cout << "Satte ny bestecase, " << cases[j].GetStringIndex("file") << " med " << score << std::endl;
		}
		else{
//			std::cout << cases[j].GetStringIndex("file") << " med " << tmp << std::endl;
		}
		std::cout << tmp << "&";
	}
	std::cout << "\\\\ \\hline" << std::endl;
	if(bestCase == 0) {
		std::cout << "something went wrong" << std::endl;
		exit(1);
	}
	
	// TODO: M REGISTRERE BILDET! 
	// KAN EVNT GJRES VED ET KALL I 
	// xml2cbr::match() til :

	//std::pair<typename ImageType::Pointer, typename ImageType::Pointer> scorePair;
	//scorePair = GetImageDistance<ImageType>(cases[i].GetStringIndex("use"), newImg, newsol, score);
	//bestImg = scorePair.first;
	//bestSol = scorePair.second;

	return *bestCase;
}


/***********************************************************************************************/













void UseThreshold2(CBRCase &incomingCase, std::string original, std::string used, float &printScore) {
	std::string saveDir = "out\\Brute2\\" + incomingCase.GetStringIndex("file") + "\\"+used+"\\";
	std::cout << saveDir << std::endl;
	std::string utDir = "";

	typedef itk::Image<int, Dimension> ImageType;
	
	std::cout << std::endl <<  "Incoming case: " << std::endl;
	incomingCase.Print();
	std::cout << std::endl;

	ImageType::Pointer orig;
	ImageType::Pointer seg;

	if(original == "old") {
	//build string with image names and get images
		std::string segString = "seg\\" + incomingCase.GetStringIndex("file") +".hdr"; 
		std::string origString = "orig\\" + incomingCase.GetStringIndex("file") + ".hdr";
		orig = GetImageFile<ImageType>(origString);
		seg = GetImageFile<ImageType>(segString);
	}
	else {
		std::cout << "casting " << std::endl;
		orig = CastImage<float, int>(bestImg);
		seg = CastImage<float, int>(bestSol);
		utDir = saveDir +"arriving.mhd";
		WriteImageFile<ImageType>(orig, utDir, &incomingCase);
	}
	//get subvolue
	//ImageType::IndexType min;
	//ImageType::IndexType max;
	//min[0] = incomingCase.GetIntIndex("Xmin");
	//min[1] = incomingCase.GetIntIndex("Ymin");
	//min[2] = incomingCase.GetIntIndex("Zmin");
	//max[0] = incomingCase.GetIntIndex("Xmax");
	//max[1] = incomingCase.GetIntIndex("Ymax");
	//max[2] = incomingCase.GetIntIndex("Zmax");

	// get subvolume. Save to disk for inspection.
	//ImageType::Pointer subimage = getSubvolume<ImageType>(orig, min, max);
	//ImageType::Pointer subsol = getSubvolume<ImageType>(seg, min, max);
	ImageType::Pointer subimage = orig;
	ImageType::Pointer subsol = seg;

//	AdaptDebug<ImageType>(subimage, orig);

//	debug
	utDir = saveDir + "subimage.mhd";
	WriteImageFile<ImageType>(subimage, utDir, &incomingCase);
		//remove the biggest region remaining after thresholding the image at 
	// argv[4]. actually set it to 255.
	ImageType::Pointer blacked = blackOut<ImageType>(subimage, incomingCase.GetIntIndex("firstThresh"));

	// write out result after largest region has been removed - debug
	utDir = saveDir + "white.mhd";
//	WriteImageFile<ImageType>(blacked, utDir, &incomingCase);

	// threshold second time
	ImageType::Pointer done = Threshold<ImageType>(subimage, incomingCase.GetIntIndex("secondThresh"));

	// only keep the largest x regions of threshold
	done = keepBiggestRegions<ImageType>(done, 1);

	// write out result of thresholding
	utDir = saveDir + "thresh-debug.mhd";
	WriteImageFile<ImageType>(done, utDir, &incomingCase);
	

	// merge thresholded subimage with large image again
	//MergeImages<ImageType>(orig, done, min);
	//utDir = saveDir + "big-overlap-debug.mhd";
	//WriteImageFile<ImageType>(orig, utDir);

	//std::cout << "Full score: ";
	//float fullsc = getScore<ImageType>(orig, seg);

	// get score of threshold.
	printScore = getScore<ImageType>(done, subsol, incomingCase.GetIntIndex("firstThresh"), incomingCase.GetIntIndex("secondThresh"));

//	PrintScoreFile(saveDir, fullsc, normsc, printScore);

	// create overlap image and write it to disk.
	subsol = createMatchImageToImage<ImageType>(done, subsol);
	utDir = saveDir + "overlap-debug.mhd";
	WriteImageFile<ImageType>(subsol, utDir, &incomingCase);

	//// create overlap image and write it to disk.
	//seg = createMatchImageToImage<ImageType>(orig, seg);
	//utDir = saveDir + "big-overlap-debug.mhd";
	//WriteImageFile<ImageType>(seg, utDir);

}


void Brute2(std::vector<CBRCase> cases) {
	float score = 0.0;
	for(int i = 8; i < 9 ; i++) {


		typedef itk::Image<float, Dimension> ImageType;
		std::string newfile = "orig\\" + cases[i].GetStringIndex("file") + ".hdr";
		ImageType::Pointer newImg = GetImageFile<ImageType>(newfile);
		std::string solstr = "seg\\" + cases[i].GetStringIndex("file") + ".hdr";
		ImageType::Pointer newsol = GetImageFile<ImageType>(solstr);

		// registrer case i med case j og threshold i
		for(int j = 0; j < cases.size(); j++) {
			if( i == j) {
				continue;
			}
			else {
				std::cout << "Registering " << cases[i].GetStringIndex("file") <<" to "<< cases[j].GetStringIndex("file") << std::endl;			
				std::pair<typename ImageType::Pointer, typename ImageType::Pointer> scorePair;
				scorePair = GetImageDistance<ImageType>(cases[j].GetStringIndex("file"), newImg, newsol, score);
				bestImg = scorePair.first;
				bestSol = scorePair.second;

				ProcessCase(cases[j], cases[i]);
			}
		} // else
		score = 0;
	}
}


void ProcessCase2(CBRCase &oldCase, CBRCase &newCase, float score) {
	
	//adapt the incoming solution.
	Adapt(oldCase, newCase);

	// process the adapted new case.
	UseThreshold2(newCase, "new", oldCase.GetStringIndex("file"), score);	

}


void PrintScoreFile(std::string dir, float full, float normal, float score) {
	std::string filename = dir +"score.txt";
	ofstream file(filename.c_str());

	std::ostringstream os1;
	std::ostringstream os2;
	std::ostringstream os3;
	os1 << full;
	os2 << normal;
	os3 << score;
	std::string ful = os1.str();
	std::string norm = os2.str();
	std::string sco = os3.str();

	//write
	file << "FinalScore: " + norm + "\n";
	file << "FirstScore: " + sco + "\n";
	file.close();
}


std::string ToString(int in) {
	std::ostringstream os;
	os << in;
	std::string ret = os.str();
	return ret;
}


void Loop(CBRCase &newCase, std::string in = "old") {
	typedef itk::Image<int,3> ImageType;

	ImageType::Pointer result;
	int cur = 0;
	int revise = 0;
	// looking for first threshold
	newCase.SetIntIndex("mode", 1);
	do{
		result = UseThreshold<ImageType>(newCase, in,cur);
		revise = Revise<ImageType>(result, newCase,cur);
		cur++;
	}while(revise != 1 && cur < 15);
	
	WriteImageFile<ImageType>(result, "out\\revise\\Final.mhd");
	exit(1);
}




/*****/
void Loop2(CBRCase newCase, CBRCase fromCase, std::string in = "old") {
	typedef itk::Image<int,3> ImageType;
	typedef itk::Image<float,3> tmpType;
	UseThreshold<ImageType>(newCase, "new", 0);

	ImageType::Pointer baseImg = CastImage<float,int>(bestImg);
//	WriteImageFile<tmpType>(bestSol, "out\\revise2\\subsolF.mhd");
	ImageType::Pointer baseSol = CastImage<float,int>(bestSol);
//	WriteImageFile<ImageType>(baseSol, "out\\revise2\\subsolF2I.mhd");

//	std::string cat = "out\\revise3\\"+newCase.GetStringIndex("file")+"\\"+fromCase.GetStringIndex("file")+"\\";
	std::string cat = "out\\revise2\\";
	std::string catImg = cat + "ArrImg.mhd";
	std::string catSol = cat + "ArrSol.mhd";
	WriteImageFile<ImageType>(baseImg, catImg);
	WriteImageFile<ImageType>(baseSol, catSol);

	std::pair<typename ImageType::Pointer, typename ImageType::Pointer> imgPair = GetSubvolumeFromCase<ImageType>(baseImg, baseSol, newCase);
	baseImg = imgPair.first,	baseSol = imgPair.second;
	int revise =0 ,cur  = 0;

//	WriteImageFile<ImageType>(baseSol, "out\\revise2\\subsol.mhd");

	// fix first threshold, set correct mode.
	newCase.SetIntIndex("mode", 1);
	do{
		std::cout << std::endl << "------------------------------NEW ROUND-------------------------------" << std::endl;
		ImageType::Pointer firstTmp = CBRCopyImage<ImageType>(baseImg);
		firstTmp = Threshold<ImageType>(firstTmp, newCase.GetIntIndex("firstThresh"));
		firstTmp = keepBiggestRegions<ImageType>(firstTmp, 1);
//		getScore<ImageType>(firstTmp, baseSol);
		std::cout << "First " << GetCentroids<ImageType>(firstTmp) << std::endl;
		revise = ReviseSecondAttempt<ImageType>(firstTmp, newCase, cur, baseSol, cat);
		cur++;
	}while(revise != 1 && cur < 100);
	revise = 0, cur = 0;
	
	// black out
	std::cout << std::endl << "------------------------------MIDDLE-------------------------------" << std::endl;
	ImageType::Pointer secondBase = CBRCopyImage<ImageType>(baseImg);
	secondBase = blackOut<ImageType>(secondBase, newCase.GetIntIndex("firstThresh"));
	WriteImageFile<ImageType>(secondBase, cat+"\\black.mhd");


	//fix second threshold
	newCase.SetIntIndex("mode", 20);
	ImageType::Pointer secondTmp;
	do{
		std::cout << std::endl << "------------------------------NEW Second ROUND-------------------------------" << std::endl;
		secondTmp = CBRCopyImage<ImageType>(secondBase);
		secondTmp = Threshold<ImageType>(secondTmp, newCase.GetIntIndex("secondThresh"));
		secondTmp = keepBiggestRegions<ImageType>(secondTmp, 1);
		getScore<ImageType>(secondTmp, baseSol);
//		GetInPlaceStatistics<ImageType>(secondTmp, secondBase);
		InsertInPlaceStatistics<ImageType>(newCase, secondTmp, secondBase);
		revise = ReviseSecondAttempt<ImageType>(secondTmp, newCase, cur, baseSol, cat);
		cur++;
	}while(revise != 1 && cur < 100);
	float FinalScore = getScore<ImageType>(secondTmp, baseSol);
	PrintScoreFile(cat, 0, FinalScore, 0);
	WriteImageFile<ImageType>(createMatchImageToImage<ImageType>(secondTmp, baseSol), cat+"Final.mhd");

//	WriteImageFile<ImageType>(secondTmp, cat+"\\Final.mhd");
}

template<class ImageType>
void InsertInPlaceStatistics(CBRCase & incomingCase, typename ImageType::Pointer segmented, typename ImageType::Pointer original) {
	std::vector<float> tmp = GetInPlaceStatistics<ImageType>(segmented, original);
	incomingCase.SetFloatIndex("InPlaceAvgTmp", tmp[0]);
	incomingCase.SetFloatIndex("InPlaceVarianceTmp", tmp[1]);
	//incomingCase.SetFloatIndex("InPlace", tmp[]);
	//incomingCase.SetFloatIndex("InPlace", tmp[]);
	//incomingCase.SetFloatIndex("InPlace", tmp[]);
	//incomingCase.SetFloatIndex("InPlace", tmp[]);
	//incomingCase.SetFloatIndex("InPlace", tmp[]);
	//incomingCase.SetFloatIndex("InPlace", tmp[]);

}

void OldSegment(CBRCase &newCase) {
	typedef itk::Image<int,3> ImageType;

	ImageType::Pointer baseImg = GetImageFile<ImageType>(newCase.GetStringIndex("file"));
	ImageType::Pointer baseSol = CastImage<float,int>(bestSol);
	std::pair<typename ImageType::Pointer, typename ImageType::Pointer> imgPair = GetSubvolumeFromCase<ImageType>(baseImg, baseSol, newCase);

	UseThreshold<ImageType>(newCase, "old", 99);
}














/*****/
void Loop3 (CBRCase newCase, CBRCase oldCase, std::string in = "old") {
	typedef itk::Image<int,3> ImageType;
	typedef itk::Image<float,3> tmpType;
	std::string cat = "out\\revise3\\"+newCase.GetStringIndex("file")+"\\"+oldCase.GetStringIndex("file")+"\\";
	//std::string cat = "out\\revise2\\";
	std::string scoreCat = "out\\Scores\\"+newCase.GetStringIndex("file")+"\\"+oldCase.GetStringIndex("file")+"\\";

	ImageType::Pointer baseImg = GetArrivedFile<ImageType>(newCase, oldCase, "ArrImg.mhd");
	ImageType::Pointer baseSol = GetArrivedFile<ImageType>(newCase, oldCase, "ArrSol.mhd");

	bestImg = CastImage<int,float>(baseImg);
	bestSol = CastImage<int,float>(baseSol);
	float firstScore = 100.0;
	UseThreshold2(newCase, "new", oldCase.GetStringIndex("file"), firstScore);
	int revise =0 ,cur  = 0;

//	WriteImageFile<ImageType>(baseSol, "out\\revise3\\subsol.mhd");

	// fix first threshold, set correct mode.
	newCase.SetIntIndex("mode", 1);
	do{
		std::cout << std::endl << "------------------------------NEW First ROUND-------------------------------" << std::endl;
		ImageType::Pointer firstTmp = CBRCopyImage<ImageType>(baseImg);
		firstTmp = Threshold<ImageType>(firstTmp, newCase.GetIntIndex("firstThresh"));
		firstTmp = keepBiggestRegions<ImageType>(firstTmp, 1);
//		getScore<ImageType>(firstTmp, baseSol);
		std::cout << "First " << GetCentroids<ImageType>(firstTmp) << std::endl;
		revise = ReviseSecondAttempt<ImageType>(firstTmp, newCase, cur, baseSol, cat);
		cur++;
	}while(revise != 1 && cur < 100);
	revise = 0, cur = 0;
	
	// black out
		std::cout << std::endl << "------------------------------MIDDLE-------------------------------" << std::endl;
	ImageType::Pointer secondBase = CBRCopyImage<ImageType>(baseImg);
	secondBase = blackOut<ImageType>(secondBase, newCase.GetIntIndex("firstThresh"));
	WriteImageFile<ImageType>(secondBase, "out\\revise2\\black.mhd");


	//fix second threshold
	newCase.SetIntIndex("mode", 20);
	ImageType::Pointer secondTmp;
	do{
		std::cout << std::endl << "------------------------------NEW Second ROUND-------------------------------" << std::endl;
		secondTmp = CBRCopyImage<ImageType>(secondBase);
		secondTmp = Threshold<ImageType>(secondTmp, newCase.GetIntIndex("secondThresh"));
		secondTmp = keepBiggestRegions<ImageType>(secondTmp, 1);
		getScore<ImageType>(secondTmp, baseSol);
//		GetInPlaceStatistics<ImageType>(secondTmp, secondBase);
		InsertInPlaceStatistics<ImageType>(newCase, secondTmp, secondBase);
		revise = ReviseSecondAttempt<ImageType>(secondTmp, newCase, cur, baseSol,cat);
		cur++;
	}while(revise != 1 && cur < 100);
	float finalScore = getScore<ImageType>(secondTmp, baseSol);
	WriteImageFile<ImageType>(createMatchImageToImage<ImageType>(secondTmp, baseSol), cat+"Final.mhd");
	PrintScoreFile(scoreCat, 0, finalScore, firstScore);
}










void BruteLast(std::vector<CBRCase> cases) {
	for(int i = 0; i < cases.size() ; i++) {

		// sammenlign med de andre casene..
		for(int j = 0; j < cases.size(); j++) {
			if( i == j) {
				continue;
			}
			else {
				ProcessCase(cases[j], cases[i]);
			}
		} // else
	}
	exit(1);
}


void QuickMatch(CBRCase newCase, std::vector<CBRCase> cases) {
	for(int j = 0; j < cases.size(); j++) {
		if( cases[j].GetStringIndex("file") == newCase.GetStringIndex("quickUse")) {
			ProcessCase(cases[j], newCase);
		}
	}	
}