#include <QWidget>
#include <QPushButton>
#include <QApplication>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QComboBox>
#include <QString>
#include <QLabel>
#include <vector>
#include <QLineEdit>
#include <QIntValidator>

#include "PLogic.h"
#include "xml2cbr.h"
#include "CBRFeature.h"
#include "CBRcase.h"

PLogic::PLogic(QWidget *parent) : QWidget(parent)
{
	cbrlogic.LoadDomainKnowledgeFromXML();
	cbrlogic.LoadCasesFromXML();
	
	//can recieve clicks
	busy = false;

	mLayout = new QVBoxLayout();
	setLayout(mLayout);
	createFinalInputWindow();
}

/**

Recieves and parses result from dropdown boxes when registering a new case.

*/
void PLogic::ddboxChanged(int a) {
	QComboBox *box = (QComboBox*)sender();

	// value of selected item
	QVariant var = box->itemData(a);
	QString varval = var.toString();

	// name of index/feature
	QString qname = box->name();
	string name = qname.toStdString();

	cbrlogic.addIndexToNewCase(name, varval.toStdString() );

//	cout << "Boxen " << name <<" ble endret til " << varval.toStdString() << endl;

}


// Slot for handling the input from the QLineEdits when inputing a new case
void PLogic::inputIntChanged() {
	QLineEdit *line = (QLineEdit*)sender();

	//get name of index and its value
	QString sValue = line->text();
	QString lName = line->name();

	cbrlogic.addIndexToNewCase(lName.toStdString(), sValue.toStdString());
}

// slot associated with the "match" button in the initial window.

void PLogic::match() {

	//already processing click?
	if(busy) {
		return;
	}
	else {
		busy = true;
		cbrlogic.match();
		//cbrlogic.saveNewCaseToXML();

		//get rid of first window. make a new one.
		//mLayout->removeWidget(firstWindow);
		//delete(firstWindow);
		//firstWindow = 0;

		//// next window, please
		//createMatchWindow();

		busy = false;
	}
}

void PLogic::quickmatch() {
	cbrlogic.quickmatch();
}


/*

Sets up the window for registration of new case.
*/

void PLogic::createInputWindow() {

	firstWindow = new QWidget(this);


	QVBoxLayout *layout = new QVBoxLayout();
	QHBoxLayout *hlayout = new QHBoxLayout();

	QPushButton * quit = new QPushButton("Quit", firstWindow);
	connect(quit, SIGNAL(clicked()), qApp, SLOT(quit()));
	
	QPushButton * match = new QPushButton("Match", firstWindow);
	connect(match, SIGNAL(clicked()), this, SLOT(match()));
	
	// iterator to go through
	map<std::string, CBRFeature> feats = cbrlogic.getFeatures();
	map<std::string, CBRFeature>::const_iterator iter;

	// add one ComboBox for each allowable index
	for(iter = feats.begin(); iter != feats.end(); ++iter) {
		CBRFeature f = iter->second;

		// for each label/input pair, place in a horizontal box.
		QHBoxLayout *ql = new QHBoxLayout();

		QLabel *label = new QLabel(QString(f.getName().c_str()), firstWindow);
		ql->addWidget(label);

		
		

		// if the input-field is numeric, get text and validate.
		if(f.isOfTypeInt()) {
			QIntValidator *ivalid = new QIntValidator(f.getMin(), f.getMax(), firstWindow);
			QLineEdit *inputline = new QLineEdit(firstWindow);
			inputline->setValidator(ivalid);
			inputline->setFixedWidth(200);
			inputline->setName(f.getName().c_str());
			ql->addWidget(inputline);

			connect(inputline, SIGNAL(editingFinished()), this, SLOT(inputIntChanged()));

			//set default value in newCase
			cbrlogic.addIndexToNewCase(f.getName(), "0");

		}
		// will be float or isOfTypeInt would kicked in
		else if(f.isNumeric()) {
			//create input field and a validator. 
			QDoubleValidator *dvalid = new QDoubleValidator(f.getMin(), f.getMax(),4, firstWindow);
			QLineEdit *inputline = new QLineEdit(firstWindow);
			inputline->setValidator(dvalid);
			inputline->setFixedWidth(200);
			inputline->setName(f.getName().c_str());
			ql->addWidget(inputline);

			connect(inputline, SIGNAL(editingFinished()), this, SLOT(inputIntChanged()));

			//set default value in newCase
			cbrlogic.addIndexToNewCase(f.getName(), "0.0");

		}
		// string, make a drop down box
		else {

			QComboBox *value = new QComboBox(firstWindow);
			value->setName(QString(f.getName().c_str()));
			value->setFixedWidth(200);
			//get name from CBRFeature
			std::vector<std::string> fvalues = f.getAllowedValues();

			// iterate through f to get allowed values
			for(unsigned int fsize = 0; fsize < fvalues.size(); fsize++ ) {
				std::string fstringtmp = fvalues.at(fsize);
				QVariant *var = new QVariant(QString(fstringtmp.c_str()));
				value->addItem(QString(fstringtmp.c_str()), *var);
			}

			connect(value, SIGNAL(currentIndexChanged(int)), this, SLOT(ddboxChanged(int)));

			// set default value in newCase
			cbrlogic.addIndexToNewCase(f.getName(), fvalues.at(0));

			// save to window
			ql->addWidget(value);
		}
		layout->addLayout(ql);
	}

	//position things
	hlayout->addWidget(quit);
	hlayout->addSpacing(30);
	hlayout->addWidget(match);


	//add the horisontal layout with the two buttons to the vertical one
	layout->addLayout(hlayout);

	//set the vertical one to apply for all firstWindow
	firstWindow->setLayout(layout);
	
	mLayout->addWidget(firstWindow);
}


void PLogic::createMatchWindow() {
	firstWindow = new QWidget(this);

	// layout of this window
	QVBoxLayout *layout = new QVBoxLayout();


	//create three labels explaining what the values are.
	QHBoxLayout *top = new QHBoxLayout();
	QLabel *labelName = new QLabel(QString("Index"), firstWindow);
	QLabel *labelNew = new QLabel(QString("From new case"), firstWindow);
	QLabel *labelBest = new QLabel(QString("From best case"), firstWindow);

	labelName->setFixedWidth(200);

	top->addWidget(labelName);
	top->addWidget(labelNew);
	top->addWidget(labelBest);

	layout->addLayout(top);


	// iterator to go through
	map<std::string, CBRFeature> feats = cbrlogic.getFeatures();
	map<std::string, CBRFeature>::const_iterator iter;
	// add one ComboBox for each allowable index
	for(iter = feats.begin(); iter != feats.end(); ++iter) {
		CBRFeature f = iter->second;

		// for each label/input pair, place in a horizontal box.
		QHBoxLayout *ql = new QHBoxLayout();

		QLabel *label = new QLabel(QString(f.getName().c_str()), firstWindow);
		label->setFixedWidth(200);
		ql->addWidget(label);

		//get value from newCase
		QLineEdit *newValue = new QLineEdit(QString(cbrlogic.getFeatureValue(f.getName(), 0).c_str()), firstWindow);
		newValue->setReadOnly(true);
		ql->addWidget(newValue);

		//get value from bestCase
		QLineEdit *bestValue = new QLineEdit(QString(cbrlogic.getFeatureValue(f.getName(), 1).c_str()), firstWindow);
		bestValue->setReadOnly(true);
		ql->addWidget(bestValue);
		
		//add to main layout
		layout->addLayout(ql);
				
	}

	// layout contains alle the little layouts.
	firstWindow->setLayout(layout);

	// add this widget  to the layout of entire window
	mLayout->addWidget(firstWindow);

	firstWindow->show();


}




/*

	New start up screen. 
	Choose one of the 20 cases to process further from a drop down box
	Contains a quit- and a match button.

*/
void PLogic::createFinalInputWindow() {
	firstWindow = new QWidget(this);


	// layouts
	QVBoxLayout *layout = new QVBoxLayout();
	QHBoxLayout *top = new QHBoxLayout();
	QHBoxLayout *bottom = new QHBoxLayout();

	// buttons and slots
	QPushButton * quit = new QPushButton("Quit", firstWindow);
	connect(quit, SIGNAL(clicked()), qApp, SLOT(quit()));
	QPushButton * match = new QPushButton("Match", firstWindow);
	connect(match, SIGNAL(clicked()), this, SLOT(match()));
	QPushButton * segment = new QPushButton("Segment", firstWindow);
	connect(segment, SIGNAL(clicked()), this, SLOT(segment()));
	QPushButton *quick = new QPushButton("QuickMatch", firstWindow);
	connect(quick, SIGNAL(clicked()), this, SLOT(quickmatch()));

	// "Choose case"
	QLabel *label = new QLabel(QString("Choose case"), firstWindow);
	top->addWidget(label);
	
	// build the drop down box with file names.
	QComboBox *value = new QComboBox(firstWindow);
	value->setName("file");
	value->setFixedWidth(100);

	//get name from CBRFeature
	std::vector<std::string> fileNames = cbrlogic.getCaseNames();

	// iterate through f to get allowed values
	for(unsigned int fsize = 0; fsize < fileNames.size(); fsize++ ) {
		std::string fstringtmp = fileNames.at(fsize);
		QVariant *var = new QVariant(QString(fstringtmp.c_str()));
		value->addItem(QString(fstringtmp.c_str()), *var);
	}

	connect(value, SIGNAL(currentIndexChanged(int)), this, SLOT(ddboxChanged(int)));

	// set default value in newCase, in case the dropdown does not change
	cbrlogic.addIndexToNewCase("file", fileNames.at(0));

	// save to window
	top->addWidget(value);
/*******************************************************************************************************************/
	QLabel *label2 = new QLabel(QString("Use as solution"), firstWindow);
	top->addWidget(label2);


	// build the drop down box with file names.
	QComboBox *value2 = new QComboBox(firstWindow);
	value2->setName("quickUse");
	value2->setFixedWidth(100);

	//get name from CBRFeature
	fileNames = cbrlogic.getCaseNames();

	// iterate through f to get allowed values
	for(unsigned int fsize = 0; fsize < fileNames.size(); fsize++ ) {
		std::string fstringtmp = fileNames.at(fsize);
		QVariant *var = new QVariant(QString(fstringtmp.c_str()));
		value2->addItem(QString(fstringtmp.c_str()), *var);
	}

	connect(value2, SIGNAL(currentIndexChanged(int)), this, SLOT(ddboxChanged(int)));

	// set default value in newCase, in case the dropdown does not change
	cbrlogic.addIndexToNewCase("quickUse", fileNames.at(0));
	top->addWidget(value2);
/*******************************************************************************************************************/


	// add the two button to the layout
	bottom->addWidget(quit);
	bottom->addWidget(segment);
	bottom->addWidget(match);
	bottom->addWidget(quick);

	// add the vertical line with the buttons to the horizontal layout
	layout->addLayout(top);
	layout->addLayout(bottom);

	firstWindow->setLayout(layout);
	mLayout->addWidget(firstWindow);
//	firstWindow->show();
}




// slot that reatcs to the segment box.
// it will keep the cbr-part out and just use the case as is, 
// mainly used in debugging and getting the histogram etc
// whitout having to input the coordinates every time.
void PLogic::segment() {
		cbrlogic.useNewCaseAsOld();
}