/*=========================================================================

  Program:   Insight Segmentation & Registration Toolkit
  Module:    $RCSfile: ImageRegistration8.cxx,v $
  Language:  C++
  Date:      $Date: 2005/12/10 17:54:30 $
  Version:   $Revision: 1.34 $

  Copyright (c) Insight Software Consortium. All rights reserved.
  See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#if defined(_MSC_VER)
#pragma warning ( disable : 4786 )
#endif





//  Software Guide : BeginCommandLineArgs
//    INPUTS: {brainweb1e1a10f20.mha}
//    INPUTS: {brainweb1e1a10f20Rot10Tx15.mha}
//    ImageRegistration8Output.mhd
//    ImageRegistration8DifferenceBefore.mhd
//    ImageRegistration8DifferenceAfter.mhd
//    OUTPUTS: {ImageRegistration8Output.png}
//    OUTPUTS: {ImageRegistration8DifferenceBefore.png}
//    OUTPUTS: {ImageRegistration8DifferenceAfter.png}
//  Software Guide : EndCommandLineArgs

// Software Guide : BeginLatex
//
// This example illustrates the use of the \doxygen{VersorRigid3DTransform}
// class for performing registration of two $3D$ images. The example code is
// for the most part identical to the code presented in
// Section~\ref{sec:RigidRegistrationIn2D}.  The major difference is that this
// example is done in $3D$. The class \doxygen{CenteredTransformInitializer} is
// used to initialize the center and translation of the transform.  The case of
// rigid registration of 3D images is probably one of the most commonly found
// cases of image registration.
//
//
// \index{itk::Versor\-Rigid3D\-Transform}
// \index{itk::Centered\-Transform\-Initializer!In 3D}
//
//
// Software Guide : EndLatex 

#include <itkImageRegistrationMethod.h>
#include "itkMeanSquaresImageToImageMetric.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkImage.h"

#include <itkNearestNeighborInterpolateImageFunction.h>
//  Software Guide : BeginLatex
//  
//  The following are the most relevant headers of this example.
//
//  \index{itk::Versor\-Rigid3D\-Transform!header}
//  \index{itk::Centered\-Transform\-Initializer!header}
// 
//  Software Guide : EndLatex 

// Software Guide : BeginCodeSnippet
#include "itkVersorRigid3DTransform.h"
#include "itkCenteredTransformInitializer.h"
// Software Guide : EndCodeSnippet

//  Software Guide : BeginLatex
//  
//  The parameter space of the \code{VersorRigid3DTransform} is not a vector
//  space, due to the fact that addition is not a closed operation in the space
//  of versor components. This precludes the use of standard gradient descent
//  algorithms for optimizing the parameter space of this transform. A special
//  optimizer should be used in this registration configuration. The optimizer
//  designed for this transform is the
//  \doxygen{VersorRigid3DTransformOptimizer}. This optimizer uses Versor
//  composition for updating the first three components of the parameters
//  array, and Vector addition for updating the last three components of the
//  parameters array~\cite{Hamilton1866,Joly1905}.
//
//  \index{itk::Versor\-Rigid3D\-Transform\-Optimizer!header}
// 
//  Software Guide : EndLatex 

// Software Guide : BeginCodeSnippet
#include "itkVersorRigid3DTransformOptimizer.h"
// Software Guide : EndCodeSnippet

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"

#include "itkResampleImageFilter.h"
#include "itkCastImageFilter.h"
#include "itkSubtractImageFilter.h"
#include "itkRescaleIntensityImageFilter.h"
#include "itkExtractImageFilter.h"
#include "itkNearestNeighborInterpolateImageFunction.h"


#include "Utility"

static float upperLimit = std::numeric_limits<float>::max();
static bool quit = false;

//  The following section of code implements a Command observer
//  that will monitor the evolution of the registration process.
//
#include "itkCommand.h"
class CommandIterationUpdate : public itk::Command 
{
public:
	typedef  CommandIterationUpdate   Self;
  typedef  itk::Command             Superclass;
  typedef itk::SmartPointer<Self>  Pointer;
  itkNewMacro( Self );
protected:
  CommandIterationUpdate() {};
public:
  typedef itk::VersorRigid3DTransformOptimizer     OptimizerType;
  typedef   const OptimizerType   *    OptimizerPointer;

  void Execute(itk::Object *caller, const itk::EventObject & event)
    {
		if(quit) {
			std::cout << "Value to high, aborting" << std::endl;
//          optimizer->SetNumberOfIterations(0);
			itk::RegularStepGradientDescentBaseOptimizer* call = (itk::RegularStepGradientDescentBaseOptimizer*) caller;
			call->StopOptimization();
		}
		Execute( (const itk::Object *)caller, event);
    }

  void Execute(const itk::Object * object, const itk::EventObject & event)
    {
      OptimizerPointer optimizer = 
        dynamic_cast< OptimizerPointer >( object );
      if( ! itk::IterationEvent().CheckEvent( &event ) )
        {
        return;
        }
      std::cout << optimizer->GetCurrentIteration() << "   ";
      std::cout << optimizer->GetValue() << "   ";
      std::cout << optimizer->GetCurrentPosition() << std::endl;
	  
	  // dont even try on the hopeless ones.
	  //if(optimizer->GetValue() > upperLimit || optimizer->GetValue() > 1500) {
		 // quit = true;
	  //}
	}
};




template<class T>
std::pair<typename T::Pointer, typename T::Pointer> Registrate( typename T::Pointer fixed, typename T::Pointer moving, 
															   typename T::Pointer solution, float &retScore /*, typename T::Pointer fullOriginal, 
															   typename T::Pointer fullSolution, typename Trans::Pointer inverseTransform*/ )
{
	std::cout << "beginning registration " << std::endl;
	quit = false;
	fixed->Update();
	moving->Update();

  if( 4 < 3 )
    {
    std::cerr << "Missing Parameters " << std::endl;
    std::cerr << "Usage: " << "Segmentation ";
    std::cerr << " fixedImageFile  movingImageFile ";
    std::cerr << " outputImagefile  [differenceBeforeRegistration] ";
    std::cerr << " [differenceAfterRegistration] ";
    std::cerr << " [sliceBeforeRegistration] ";
    std::cerr << " [sliceAfterRegistration] "<< std::endl;
    exit(0);
    }
  
  const    int			   Dimension = 3;
  typedef  float           PixelType;

  typedef itk::Image< PixelType, Dimension >  FixedImageType;
  typedef itk::Image< PixelType, Dimension >  MovingImageType;
  //typedef itk::Image< PixelType, Dimension >  FixedImageType;
  //typedef itk::Image< PixelType, Dimension >  MovingImageType;


  //  Software Guide : BeginLatex
  //  
  //  The Transform class is instantiated using the code below. The only
  //  template parameter to this class is the representation type of the
  //  space coordinates.
  //
  //  \index{itk::Versor\-Rigid3D\-Transform!Instantiation}
  //
  //  Software Guide : EndLatex 

  // Software Guide : BeginCodeSnippet

  //originally double
  typedef itk::VersorRigid3DTransform< double > TransformType;
  // Software Guide : EndCodeSnippet



  typedef itk::VersorRigid3DTransformOptimizer           OptimizerType;

  typedef itk::MeanSquaresImageToImageMetric< 
                                    FixedImageType, 
                                    MovingImageType >    MetricType;

  typedef itk:: LinearInterpolateImageFunction< 
                                    MovingImageType,
                                    double          >    InterpolatorType;
  typedef itk::NearestNeighborInterpolateImageFunction< 
                                    MovingImageType,
                                    double          >    InterpolatorType1;

  typedef itk::ImageRegistrationMethod< 
                                    FixedImageType, 
                                    MovingImageType >    RegistrationType;

  MetricType::Pointer         metric        = MetricType::New();
  OptimizerType::Pointer      optimizer     = OptimizerType::New();
  InterpolatorType::Pointer   interpolator  = InterpolatorType::New();
  InterpolatorType1::Pointer   interpolator1  = InterpolatorType1::New();
  RegistrationType::Pointer   registration  = RegistrationType::New();
  

  registration->SetMetric(        metric        );
  registration->SetOptimizer(     optimizer     );
  registration->SetInterpolator(  interpolator  );


  //  Software Guide : BeginLatex
  //
  //  The transform object is constructed below and passed to the registration
  //  method.
  //
  //  \index{itk::Versor\-Rigid3D\-Transform!New()}
  //  \index{itk::Versor\-Rigid3D\-Transform!Pointer}
  //  \index{itk::Registration\-Method!SetTransform()}
  //
  //  Software Guide : EndLatex 

  // Software Guide : BeginCodeSnippet
  TransformType::Pointer  transform = TransformType::New();
  registration->SetTransform( transform );
  // Software Guide : EndCodeSnippet

  typedef itk::ImageFileReader< FixedImageType  > FixedImageReaderType;
  typedef itk::ImageFileReader< MovingImageType > MovingImageReaderType;

  FixedImageReaderType::Pointer  fixedImageReader  = FixedImageReaderType::New();
  MovingImageReaderType::Pointer movingImageReader = MovingImageReaderType::New();

  //fixedImageReader->SetFileName(  argv[1] );
  //movingImageReader->SetFileName( argv[13] );

  registration->SetFixedImage(    fixed   );
  registration->SetMovingImage(   moving   );
//  fixedImageReader->Update();

  registration->SetFixedImageRegion( 
     fixed->GetBufferedRegion() );


  //  Software Guide : BeginLatex
  //  
  //  The input images are taken from readers. It is not necessary here to
  //  explicitly call \code{Update()} on the readers since the
  //  \doxygen{CenteredTransformInitializer} will do it as part of its
  //  computations. The following code instantiates the type of the
  //  initializer. This class is templated over the fixed and moving image type
  //  as well as the transform type. An initializer is then constructed by
  //  calling the \code{New()} method and assigning the result to a smart
  //  pointer.
  //
  // \index{itk::Centered\-Transform\-Initializer!Instantiation}
  // \index{itk::Centered\-Transform\-Initializer!New()}
  // \index{itk::Centered\-Transform\-Initializer!SmartPointer}
  //
  //  Software Guide : EndLatex 

  // Software Guide : BeginCodeSnippet
  typedef itk::CenteredTransformInitializer< TransformType, 
                                             FixedImageType, 
                                             MovingImageType 
                                                 >  TransformInitializerType;

  TransformInitializerType::Pointer initializer = 
                                          TransformInitializerType::New();
  // Software Guide : EndCodeSnippet


  //  Software Guide : BeginLatex
  //
  //  The initializer is now connected to the transform and to the fixed and
  //  moving images.
  //
  //  Software Guide : EndLatex 


  // Software Guide : BeginCodeSnippet
  initializer->SetTransform(   transform );
  initializer->SetFixedImage(  fixed );
  initializer->SetMovingImage( moving );
  // Software Guide : EndCodeSnippet

  //  Software Guide : BeginLatex
  //
  //  The use of the geometrical centers is selected by calling
  //  \code{GeometryOn()} while the use of center of mass is selected by
  //  calling \code{MomentsOn()}.  Below we select the center of mass mode.
  //
  //  \index{Centered\-Transform\-Initializer!MomentsOn()}
  //  \index{Centered\-Transform\-Initializer!GeometryOn()}
  //
  //  Software Guide : EndLatex 

  // Software Guide : BeginCodeSnippet
  initializer->MomentsOn();
  // Software Guide : EndCodeSnippet


  //  Software Guide : BeginLatex
  //
  //  Finally, the computation of the center and translation is triggered by
  //  the \code{InitializeTransform()} method. The resulting values will be
  //  passed directly to the transform.
  //
  //  Software Guide : EndLatex 


  // Software Guide : BeginCodeSnippet
  initializer->InitializeTransform();
  // Software Guide : EndCodeSnippet


  //  Software Guide : BeginLatex
  //  
  //  The rotation part of the transform is initialized using a
  //  \doxygen{Versor} which is simply a unit quaternion.  The
  //  \code{VersorType} can be obtained from the transform traits. The versor
  //  itself defines the type of the vector used to indicate the rotation axis.
  //  This trait can be extracted as \code{VectorType}. The following lines
  //  create a versor object and initialize its parameters by passing a
  //  rotation axis and an angle.
  //
  //  Software Guide : EndLatex 

  // Software Guide : BeginCodeSnippet
  typedef TransformType::VersorType  VersorType;
  typedef VersorType::VectorType     VectorType;

  VersorType     rotation;
  VectorType     axis;
  
  axis[0] = 0.0;
  axis[1] = 0.0;
  axis[2] = 1.0;

  const double angle = 0;

  rotation.Set(  axis, angle  );

  transform->SetRotation( rotation );
  // Software Guide : EndCodeSnippet


  //  Software Guide : BeginLatex
  //  
  //  We now pass the parameters of the current transform as the initial
  //  parameters to be used when the registration process starts. 
  //
  //  Software Guide : EndLatex 

  // Software Guide : BeginCodeSnippet
  registration->SetInitialTransformParameters( transform->GetParameters() );
  // Software Guide : EndCodeSnippet

  typedef OptimizerType::ScalesType       OptimizerScalesType;
  OptimizerScalesType optimizerScales( transform->GetNumberOfParameters() );
  const double translationScale = 1.0 / 1000.0;

  optimizerScales[0] = 1.0;
  optimizerScales[1] = 1.0;
  optimizerScales[2] = 1.0;
  optimizerScales[3] = translationScale;
  optimizerScales[4] = translationScale;
  optimizerScales[5] = translationScale;

  optimizer->SetScales( optimizerScales );

  optimizer->SetMaximumStepLength( 0.2000  ); 
  optimizer->SetMinimumStepLength( 0.0001 );

  optimizer->SetNumberOfIterations( 200 );


  // Create the Command observer and register it with the optimizer.
  //
  CommandIterationUpdate::Pointer observer = CommandIterationUpdate::New();
  optimizer->AddObserver( itk::IterationEvent(), observer );


  try 
    { 
    registration->StartRegistration(); 
    } 
  catch( itk::ExceptionObject & err ) 
    { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    exit(1);
    } 
  
  OptimizerType::ParametersType finalParameters = 
                    registration->GetLastTransformParameters();


  const double versorX              = finalParameters[0];
  const double versorY              = finalParameters[1];
  const double versorZ              = finalParameters[2];
  const double finalTranslationX    = finalParameters[3];
  const double finalTranslationY    = finalParameters[4];
  const double finalTranslationZ    = finalParameters[5];

  const unsigned int numberOfIterations = optimizer->GetCurrentIteration();

  const double bestValue = optimizer->GetValue();


  // Print out results
  //
  //std::cout << std::endl << std::endl;
  //std::cout << "Result = " << std::endl;
  //std::cout << " versor X      = " << versorX  << std::endl;
  //std::cout << " versor Y      = " << versorY  << std::endl;
  //std::cout << " versor Z      = " << versorZ  << std::endl;
  //std::cout << " Translation X = " << finalTranslationX  << std::endl;
  //std::cout << " Translation Y = " << finalTranslationY  << std::endl;
  //std::cout << " Translation Z = " << finalTranslationZ  << std::endl;
  //std::cout << " Iterations    = " << numberOfIterations << std::endl;
  //std::cout << " Metric value  = " << bestValue          << std::endl;



  // Software Guide : BeginCodeSnippet
  transform->SetParameters( finalParameters );

  TransformType::MatrixType matrix = transform->GetRotationMatrix();
  TransformType::OffsetType offset = transform->GetOffset();


  typedef itk::ResampleImageFilter< 
                            MovingImageType, 
                            FixedImageType >    ResampleFilterType;

  TransformType::Pointer finalTransform = TransformType::New();

  finalTransform->SetCenter( transform->GetCenter() );
  finalTransform->SetParameters( finalParameters );


// TransformType::Pointer inverseTransform2 = TransformType::New();

 //get inverse
 //finalTransform->GetInverse(inverseTransform); 
// std::cout << "Final distance: " << optimizer->GetValue() << std::endl;
	retScore = optimizer->GetValue();
	
	// try to cut hopeless cases early.
	if(retScore < upperLimit-600) {
		upperLimit = retScore + 600;
	}

// resample original input
  ResampleFilterType::Pointer resampler = ResampleFilterType::New();
//resample solution input
  ResampleFilterType::Pointer resamplerSol = ResampleFilterType::New();

  // initialize transforms.
  resampler->SetTransform( finalTransform );
  resampler->SetInput( moving );
  resamplerSol->SetTransform( finalTransform);
  resamplerSol->SetInput( solution );

  FixedImageType::Pointer fixedImage = fixed;
	// set up for resampler for original image
  resampler->SetSize(    fixedImage->GetLargestPossibleRegion().GetSize() );
  resampler->SetOutputOrigin(  fixedImage->GetOrigin() );
  resampler->SetOutputSpacing( fixedImage->GetSpacing() );
  resampler->SetDefaultPixelValue( 0 );
//  resampler->SetInterpolator(interpolator1);

  // set up for resampler for solution, copied from above
  resamplerSol->SetSize(    fixedImage->GetLargestPossibleRegion().GetSize() );
  resamplerSol->SetOutputOrigin(  fixedImage->GetOrigin() );
  resamplerSol->SetOutputSpacing( fixedImage->GetSpacing() );
  resamplerSol->SetDefaultPixelValue( 0 );
  resamplerSol->SetInterpolator(interpolator1);


  typedef  float  OutputPixelType;

  typedef itk::Image< OutputPixelType, Dimension > OutputImageType;
  
  typedef itk::CastImageFilter< 
                        FixedImageType,
                        OutputImageType > CastFilterType;
                    
  typedef itk::ImageFileWriter< OutputImageType >  WriterType;


  WriterType::Pointer      writer =  WriterType::New();
  CastFilterType::Pointer  caster =  CastFilterType::New();


//  writer->SetFileName( argv[3] );


//  caster->SetInput( resampler->GetOutput() );
//  writer->SetInput( caster->GetOutput()   );
  
	

 // try{writer->Update(); }
 // catch(itk::ExceptionObject &exp) {
	//  std::cerr << exp;
	//exit(1);
 // }

  try{ 
	  resampler->Update(); 
	  resamplerSol->Update();
  }
  catch(itk::ExceptionObject &exp) { std::cerr << exp; exit(1); }

  typedef std::pair<T::Pointer, T::Pointer> PairType;
  PairType pair;
  pair.first = resampler->GetOutput();
  pair.second = resamplerSol->GetOutput();

  return pair;

}









template<class T, class Trans>
typename T::Pointer unRegister(typename T::Pointer image,   typename Trans::Pointer Transform) {

	std::cout << "Starting reverse transformation "<< std::endl;
	typedef itk::ResampleImageFilter<T, T>    ResampleFilterType;
	typedef itk::NearestNeighborInterpolateImageFunction<T, double> InterpolatorType;
	
	
	
	//itk::ImageFileWriter<T>::Pointer writer = itk::ImageFileWriter<T>::New();
	//writer->SetInput(image);
	//writer->SetFileName("unRegImage.mhd");
	//writer->Update();

	ResampleFilterType::Pointer resampler = ResampleFilterType::New();
	InterpolatorType::Pointer inter = InterpolatorType::New();

	//resampler::InterpolatorType interpolator;
	//interpolator->

	resampler->SetInterpolator(inter);
	resampler->SetTransform( Transform );
	resampler->SetInput( image );
	resampler->SetSize(    image->GetLargestPossibleRegion().GetSize() );
	resampler->SetOutputOrigin(  image->GetOrigin() );
	resampler->SetOutputSpacing( image->GetSpacing() );
	resampler->SetDefaultPixelValue( 0 );


	T::Pointer ret = resampler->GetOutput();
	
	try{
		ret->Update();
	}
	catch(itk::ExceptionObject &exp) {std::cerr << exp; exit(1);}

	//writer->SetInput(ret);
	//writer->SetFileName("unRegPrImage.mhd");
	//writer->Update();
	std::cout << "Finished reverse "<< std::endl;
	
	return ret;
}
